try:
    import weightwatcher as ww
    WEIGHTWATCHER_PRESENT = True
except:
    WEIGHTWATCHER_PRESENT = False

try:
    import torch
    TORCH_PRESENT = True
except:
    TORCH_PRESENT = False


from uptrain.v0.core.classes.monitors import AbstractCheck


class Finetune(AbstractCheck):

    def base_init(self, fw, check):
        self.optimizer = check["optimizer"]
        self.train_dataloader = check["train_dataloader"]
        self.val_dataloader = check["val_dataloader"]
        self.layers = check["layers"]
        self.epochs = check.get("epochs", 1)
        self.device = check.get("device", "cpu")
        self.dashboard_name = "Finetuning Statistics"
        self.is_automated = check.get("is_automated", True)
        self.base_lr = check.get("base_lr", 2e-5)

    def getCustomLRParms(self, model):
        # TODO: Make 'post_bert' generic
        finetuned_names = [k for (k, v) in model.named_parameters() if 'post_bert' in k or 'classifier' in k]
        new_params= [v for k, v in model.named_parameters() if k in finetuned_names]
        pretrained = [v for k, v in model.named_parameters() if k not in finetuned_names]
        return new_params, pretrained

    def getTorchOptimizer(self, model, multipleFactor=1):
        if self.is_automated:
            new_params, pretrained = self.getCustomLRParms(model)
            optimizer_grouped_parameters = [
                {'params': new_params, 'lr': multipleFactor * self.base_lr},
                {'params': pretrained, 'lr': self.base_lr}
            ]
            optimizer = self.optimizer(optimizer_grouped_parameters)
            return optimizer
        else:
            return self.optimizer(model.parameters(), lr=self.base_lr)

    def base_check(self, inputs, outputs, gts=None, extra_args={}):
        model = inputs['model'][0]

        for name, param in model.named_parameters():
            if 'classifier' in name or 'post_bert' in name:
                print(name)
                param.requires_grad = True
            else:
                param.requires_grad = False


        optimizer = self.getTorchOptimizer(model)
        watcher = ww.WeightWatcher(model=model)
        model.train()

        device = self.device
        num_all_points = 0
        num_all_val_points = 0

        if not WEIGHTWATCHER_PRESENT:
            raise Exception("WeightWatcher is not installed. Please install it using 'pip install weightwatcher'")
        if not TORCH_PRESENT:
            raise Exception("PyTorch is not installed. Please install it using 'pip install torch'")
        for epoch in range(self.epochs):
            tr_loss = 0
            val_loss = 0
            nb_tr_examples, nb_tr_steps = 0, 0
            nb_val_examples, nb_val_steps = 0, 0
            for step, batch in enumerate(self.train_dataloader):

                a = watcher.analyze(layers=self.layers)
                a_list = a['alpha'].tolist()
                if self.is_automated:
                    if a_list[0] > 8:
                        optimizer = self.getTorchOptimizer(model, 5)

                alpha_avg = sum(a_list)/len(a_list)
                self.log_handler.add_scalars(
                    "alpha_avg",
                    {"y_alpha_avg": alpha_avg},
                    num_all_points,
                    self.dashboard_name,
                )

                batch = tuple(t.to(device) for t in batch)
                b_input_ids, b_input_mask, b_labels = batch
                b_labels = b_labels.to(device)
                b_input_ids = b_input_ids.to(device)
                b_input_mask = b_input_mask.to(device)
                
                optimizer.zero_grad()
                # Forward pass
                train_output = model(b_input_ids, 
                                    attention_mask = b_input_mask, 
                                        token_type_ids = None,
                                    labels = b_labels)
                # Backward pass
                train_output.loss.backward()
                optimizer.step()
                # Update tracking variables
                tr_loss += train_output.loss.item()
                nb_tr_examples += b_input_ids.size(0)
                num_all_points += b_input_ids.size(0)
                nb_tr_steps += 1

            self.log_handler.add_scalars(
                    "training_loss",
                    {"y_train_loss": tr_loss/nb_tr_steps},
                    num_all_points,
                    self.dashboard_name,
                )

            val_loss = 0
            nb_val_examples = 0
            for batch in self.val_dataloader:
                batch = tuple(t.to(device) for t in batch)
                b_input_ids, b_input_mask, b_labels = batch
                with torch.no_grad():
                    # Forward pass
                    eval_output = model(b_input_ids.to(device), 
                                        token_type_ids = None, 
                                        attention_mask = b_input_mask.to(device),
                                            labels = b_labels.to(device)
                                        )
                val_loss += eval_output.loss.item()
                nb_val_examples += b_input_ids.size(0)
                num_all_val_points += b_input_ids.size(0)
                nb_val_steps += 1

            self.log_handler.add_scalars(
                "validation_loss",
                {"y_val_loss": val_loss/nb_val_steps},
                num_all_points,
                self.dashboard_name,
            )



"""
Implement checks to return the input dataset as it is.

This module provides the `ColumnExpand` class, which returns the input DataFrame as it is without any modifications. It can be used to pass the data through unchanged in a pipeline.

"""

from __future__ import annotations
import typing as t

import numpy as np
from loguru import logger
from pydantic import Field
import polars as pl

if t.TYPE_CHECKING:
    from uptrain.framework import Settings
from uptrain.operators.base import *

@register_op
class Table(OpBaseModel):
    """
    Operator to generate a table.

    Attributes:
        props (dict): Additional properties to pass to the Table constructor.
        title (str): The title of the chart.

    """

    props: dict = Field(default_factory=dict)
    title: str = ""

    kind = "table"

    def setup(self, settings: Settings):
        return self

    def run(self, data: pl.DataFrame) -> TYPE_TABLE_OUTPUT:
        return {"output": None}


@register_op
class ColumnExpand(ColumnOp):
    """
    Table operation to return the input DataFrame as it is without any modifications.

    Attributes:
        col_out_names (list[str]): The names of the output columns. Must be the same length as `col_vals`.
        col_vals (list[Any]): The values for the output columns. Must be the same length as `col_out_names`.

    Returns:
        dict: A dictionary containing the input DataFrame.

    Example:
        ```
        from uptrain.operators import ColumnExpand
        df = pl.DataFrame({
            "column1": [1, 2, 3],
            "column2": ["A", "B", "C"]
        })

        # Create an instance of the ColumnExpand class
        expand_op = ColumnExpand(
                        col_out_names=["column1", "column2"],
                        col_vals=[df["column1"], df["column2"]]
                    )

        # Run the expand operation
        output_df = expand_op.run(df)["output"]

        # Print the output DataFrame
        print(output_df)
        ```

    Output:
        ```
        shape: (3, 2)
        ┌─────────┬─────────┐
        │ column1 ┆ column2 │
        │ ---     ┆ ---     │
        │ i64     ┆ str     │
        ╞═════════╪═════════╡
        │ 1       ┆ A       │
        │ 2       ┆ B       │
        │ 3       ┆ C       │
        └─────────┴─────────┘
        ```

    """

    col_out_names: list[str]
    col_vals: list[t.Any]

    def setup(self, settings: Settings):
        assert len(self.col_out_names) == len(self.col_vals)
        return self

    def run(self, data: pl.DataFrame) -> TYPE_TABLE_OUTPUT:
        out = data.with_columns(
            [
                pl.lit(self.col_vals[idx]).alias(self.col_out_names[idx])
                for idx in range(len(self.col_out_names))
            ]
        )
        return {"output": out}


@register_op
class ColumnComparison(ColumnOp):
    """
    Column operation to add a new boolean column which represents if the given columns are row-wise same or not

    Attributes:
        col_in_1 (str): The name of the 1st input column to be compared row-wise.
        col_in_2 (str): The name of the 2nd input column to be compared row-wise.
        col_out (str): The name of the output boolean column 

    Returns:
        dict: A dictionary containing the input DataFrame.

    Example:
        ```
        from uptrain.operators import ColumnComparison
        df = pl.DataFrame({
            "column1": [1, 2, 3],
            "column2": [1, 2, 1],
            "column3": [2, 2, 4]
        })

        # Create an instance of the ColumnComparison class
        compare_op = ColumnComparison(
                        col_in_1="column1",
                        col_in_2="column2",
                        col_out="is_equal"
                    )

        # Run the compare operation
        output_df = compare_op.run(df)["output"]

        # Print the output DataFrame
        print(output_df)
        ```

    Output:
        ```
        shape: (3, 2)
        ┌─────────┬─────────┐─────────┐─────────┐
        │ column1 ┆ column2 ┆ column3 ┆is_equal │
        │ ---     ┆ ---     ┆ ---     ┆ ---     │
        │ i64     ┆ i64     ┆ i64     ┆ bool    │
        ╞═════════╪═════════╪═════════╪═════════╡
        │ 1       ┆ 1       ┆ 2       ┆ True    │
        │ 2       ┆ 2       ┆ 2       ┆ True    │
        │ 3       ┆ 1       ┆ 4       ┆ False   │
        └─────────┴─────────┴─────────┴─────────┘
        ```

    """

    col_in_1: str
    col_in_2: str
    col_out: str

    def setup(self, settings: Settings):
        return self

    def run(self, data: pl.DataFrame) -> TYPE_TABLE_OUTPUT:
        return {"output": data.with_columns(data.select(
            pl.col(self.col_in_1) ==
            pl.col(self.col_in_2)
        )['Output'].alias(self.col_out))}


@register_op
class ColumnReduce(ColumnOp):
    """
    Operator to drop a list of columns from the input dataframe.

    Attributes:
        col_drop_names (list[str]): The names of the columns to be dropped.

    Returns:
        dict: A dictionary containing the input DataFrame.

    Example:
        ```
        from uptrain.operators import ColumnReduce
        df = pl.DataFrame({
            "column1": [1, 2, 3],
            "column2": ["A", "B", "C"]
        })

        # Create an instance of the ColumnReduce class
        reduce_op = ColumnReduce(
                        col_drop_names=["column2"],
                    )

        # Run the reduce operation
        output_df = reduce_op.run(df)["output"]

        # Print the output DataFrame
        print(output_df)
        ```

    Output:
        ```
        shape: (3, 2)
        ┌─────────┐
        │ column1 ┆
        │ ---     ┆
        │ i64     ┆
        ╞═════════╡
        │ 1       ┆
        │ 2       ┆
        │ 3       ┆
        └─────────┘
        ```

    """

    col_drop_names: list[str]

    def setup(self, settings: Settings):
        return self

    def run(self, data: pl.DataFrame) -> TYPE_TABLE_OUTPUT:
        out = data.drop(self.col_drop_names)
        return {"output": out}



"""
Implement operators to evaluate question-response-guideline and question-response datapoints for a 
LLM-powered agent pipeline. 
"""

from __future__ import annotations
import typing as t

from loguru import logger
import polars as pl

if t.TYPE_CHECKING:
    from uptrain.framework import Settings
from uptrain.operators.base import *


@register_op
class GuidelineAdherenceScore(ColumnOp):
    """
    Grade if the LLM agent follows the given guideline or not.

     Attributes:
        col_question (str): Column name for the stored questions
        col_response (str): Column name for the stored responses
        guideline (str): Guideline to be followed
        guideline_name (str): Name the given guideline to be used in the score column name
        response_schema (str | None): Schema of the response if it is in form of a json string
        col_out (str): Column name to output scores

    Raises:
        Exception: Raises exception for any failed evaluation attempts

    """

    col_question: str = "question"
    col_response: str = "response"
    guideline: str
    guideline_name: str = "guideline"
    response_schema: t.Union[str, None] = None
    col_out: str  = "score_guideline_adherence"

    def setup(self, settings: t.Optional[Settings] = None):
        from uptrain.framework.remote import APIClient

        assert settings is not None
        self._api_client = APIClient(settings)
        return self

    def run(self, data: pl.DataFrame) -> TYPE_TABLE_OUTPUT:
        data_send = [
            {
                "question": row[self.col_question],
                "response": row[self.col_response],
            }
            for row in data.to_dicts()
        ]
        try:
            results = self._api_client.evaluate(
                "GuidelineAdherence", data_send, {
                    "guideline": self.guideline,
                    "guideline_name": self.guideline_name,
                    "response_schema": self.response_schema
                })
                
        except Exception as e:
            logger.error(f"Failed to run evaluation for `GuidelineAdherenceScore`: {e}")
            raise e

        assert results is not None
        return {"output": data.with_columns(pl.from_dicts(results).rename({f"score_{self.guideline_name}_adherence": self.col_out}))}

@register_op
class PromptInjectionScore(ColumnOp):
    """
    Grade if the LLM agent outputs the system prompt or not.

     Attributes:
        col_question (str): Column name for the stored questions
        col_response (str): Column name for the stored responses
        col_out (str): Column name to output scores
    Raises:
        Exception: Raises exception for any failed evaluation attempts

    """

    col_question: str = "question"
    col_response: str = "response"
    col_out: str  = "score_prompt_injection"

    def setup(self, settings: t.Optional[Settings] = None):
        from uptrain.framework.remote import APIClient

        assert settings is not None
        self._api_client = APIClient(settings)
        return self

    def run(self, data: pl.DataFrame) -> TYPE_TABLE_OUTPUT:
        data_send = [
            {
                "question": row[self.col_question],
                "response": row[self.col_response],
            }
            for row in data.to_dicts()
        ]
        try:
            results = self._api_client.evaluate(
                "prompt_injection", data_send)
                
        except Exception as e:
            logger.error(f"Failed to run evaluation for `PromptInjectionScore`: {e}")
            raise e

        assert results is not None
        return {"output": data.with_columns(pl.from_dicts(results).rename({f"score_prompt_injection": self.col_out}))}



import uptrain.v0 as v0
from test_helpers import KpsDataset, get_data_from_remote
import torch
from test_helpers import (
    body_length_signal,
    train_model_torch,
    get_accuracy_torch,
    BinaryClassification,
)


# if __name__ == "__main__":
def test_data_drift():
    get_data_from_remote()

    inference_batch_size = 16

    # Training data
    training_file = "data/training_data.json"

    # A testing dataset to evaluate model performance
    golden_testing_file = "data/golden_testing_data.json"

    # The data-points which the models sees in production
    real_world_test_cases = "data/real_world_testing_data.json"

    # To annotate the collected data points, we extract the ground truth from a master annotation file
    # (we can also import from any other annotation pipelines such as an annotation job on Mechanical turk).
    annotation_args = {"master_file": "data/master_annotation_data.json"}

    train_model_torch(training_file, "version_0")

    check1 = {
        "type": v0.Monitor.DATA_DRIFT,
        "reference_dataset": training_file,
        "is_embedding": True,
        "measurable_args": {
            "type": v0.MeasurableType.INPUT_FEATURE,
            "feature_name": "kps",
        },
    }

    check2 = {
        "type": v0.Monitor.DATA_DRIFT,
        "reference_dataset": training_file,
        "save_edge_cases": False,
        "measurable_args": {
            "type": v0.MeasurableType.SCALAR_FROM_EMBEDDING,
            "idx": 0,
            "extract_from": {
                "type": v0.MeasurableType.INPUT_FEATURE,
                "feature_name": "kps",
            },
        },
    }

    check3 = {
        "type": v0.Monitor.DATA_DRIFT,
        "reference_dataset": training_file,
        "is_embedding": False,
        "save_edge_cases": False,
        "measurable_args": {
            "type": v0.MeasurableType.CUSTOM,
            "signal_formulae": v0.Signal("Body Length", body_length_signal),
        },
    }

    checks = [check1, check2, check3]

    # Define the training pipeline to annotate data and retrain the model automatically
    training_args = {
        "annotation_method": {
            "method": v0.AnnotationMethod.MASTER_FILE,
            "args": annotation_args,
        },
        "training_func": train_model_torch,
        "orig_training_file": training_file,
    }

    # Define evaluation pipeline to test retrained model against original model
    evaluation_args = {
        "inference_func": get_accuracy_torch,
        "golden_testing_dataset": golden_testing_file,
    }

    cfg = {
        "checks": checks,
        "training_args": training_args,
        "evaluation_args": evaluation_args,
        # Retrain when 200 datapoints are collected in the retraining dataset
        "retrain_after": 200,
        # A local folder to store the retraining dataset
        "retraining_folder": "uptrain_smart_data_data_drift",
    }

    framework = v0.Framework(cfg)

    inference_batch_size = 16
    model_dir = "trained_models_torch/"
    model_save_name = "version_0"
    real_world_dataset = KpsDataset(
        real_world_test_cases, batch_size=inference_batch_size, is_test=True
    )
    model = BinaryClassification()
    model.load_state_dict(torch.load(model_dir + model_save_name))
    model.eval()

    for i, elem in enumerate(real_world_dataset):
        # Do model prediction
        inputs = {"kps": elem[0]["kps"], "id": elem[0]["id"]}
        x_test = torch.tensor(inputs["kps"]).type(torch.float)
        test_logits = model(x_test).squeeze()
        preds = torch.round(torch.sigmoid(test_logits)).detach().numpy()

        # Log model inputs and outputs to the uptrain Framework to monitor data drift
        idens = framework.log(inputs=inputs, outputs=preds)

        # Retrain only once
        if framework.version > 1:
            break



"""
Implements operator to calculate the matching score between llm-generated response and ground_truth response.
"""

from __future__ import annotations
import typing as t

from loguru import logger
import polars as pl

if t.TYPE_CHECKING:
    from uptrain.framework import Settings
from uptrain.operators.base import *


@register_op
class ResponseMatchingScore(ColumnOp):
    """
    Operator to compare the llm-generated text with the gold response using the defined score metric.

     Attributes:
        col_response (str): Column name for the llm generated responses
        col_ground_truth (str): Column name for the ground truth responses
        method (str): (Literal["rouge", "exact", "llm"]): Method to calculate the score

    Raises:
        Exception: Raises exception for any failed evaluation attempts

    """

    col_response: str = "response"
    col_ground_truth: str = "ground_truth"
    method: str = t.Literal["exact", "rouge", "llm"]
    col_out: str = "score_response_match"

    def setup(self, settings: t.Optional[Settings] = None):
        from uptrain.framework.remote import APIClient
        assert settings is not None
        if self.method not in ["exact", "rouge", "llm"]:
            raise Exception(f"Metric: {self.method} is not supported yet.")
        self._api_client = APIClient(settings)
        return self

    def run(self, data: pl.DataFrame) -> TYPE_TABLE_OUTPUT:
        data_send = [
            {
                "response": row[self.col_response],
                "ground_truth": row[self.col_ground_truth],
            }
            for row in data.to_dicts()
        ]
        try:
            results = self._api_client.evaluate(
                "ResponseMatching", data_send, {
                    "type": self.method            
                })

        except Exception as e:
            logger.error(f"Failed to run evaluation for `ResponseMatchingScore`: {e}")
            raise e

        assert results is not None
        return {"output": data.with_columns(pl.from_dicts(results).rename({"score_response_match": self.col_out}))}



@register_op
class ValidResponseScore(ColumnOp):
    """
    Operator to check if the response is valid is not.
     Attributes:
        col_response (str): Column name for the responses
        col_out (str): Column name for the output score

    Raises:
        Exception: Raises exception for any failed evaluation attempts

    """

    col_response: str = "response"
    col_out: str = "score_valid_response"

    def setup(self, settings: t.Optional[Settings] = None):
        from uptrain.framework.remote import APIClient
        assert settings is not None
        self._api_client = APIClient(settings)
        return self

    def run(self, data: pl.DataFrame) -> TYPE_TABLE_OUTPUT:
        data_send = [
            {
                "response": row[self.col_response],
            }
            for row in data.to_dicts()
        ]
        try:
            results = self._api_client.evaluate("valid_response", data_send)

        except Exception as e:
            logger.error(f"Failed to run evaluation for `ResponseMatchingScore`: {e}")
            raise e

        assert results is not None
        return {"output": data.with_columns(pl.from_dicts(results).rename({"score_valid_response": self.col_out}))}



from typing import Any

from uptrain.v0.core.classes.measurables import (
    Measurable,
    InputFeatureMeasurable,
)


class ScalarFromEmbeddingMeasurable(Measurable):
    """Class that extracts a scalar from a measurable embedding."""

    def __init__(self, framework, idx, extract_from_args) -> None:
        super().__init__(framework)
        self.idx = idx
        self.extract_from = InputFeatureMeasurable(
            framework, extract_from_args["feature_name"]
        )

    def _compute(self, inputs=None, outputs=None, gts=None, extra=None) -> Any:
        return self.extract_from._compute(
            inputs=inputs, outputs=outputs, gts=gts, extra=extra
        )[self.idx]

    def col_name(self) -> str:
        return f"Scalar [Index: {str(self.idx)}] from Embedding: {str(self.extract_from.col_name())}"

    # TODO: Decommission and find a generic way
    def extract_val_from_training_data(self, x):
        return self.extract_from.extract_val_from_training_data(x)[self.idx]



"""
This module implements the ConceptDrift operator and its associated executor for detecting concept drift
using the DDM (Drift Detection Method) or ADWIN (Adaptive Windowing) algorithm.

The module provides the following classes:

- ParamsDDM: Parameters for the DDM algorithm.
- ParamsADWIN: Parameters for the ADWIN algorithm.
- ConceptDrift: Operator for detecting concept drift.
- ConceptDriftExecutor: Executor for the ConceptDrift operator.

"""

from __future__ import annotations
import typing as t

from loguru import logger
from pydantic import root_validator
import polars as pl

if t.TYPE_CHECKING:
    from uptrain.framework import Settings
from uptrain.operators.base import *
from uptrain.utilities import lazy_load_dep

drift = lazy_load_dep("river.drift", "river")


class ParamsDDM(OpBaseModel):
    """
    Parameters for the DDM (Drift Detection Method) algorithm.

    Attributes:
        warm_start (int): The number of instances required before any drift detection.
        warning_threshold (float): The warning threshold value for the drift detection.
        drift_threshold (float): The alarm threshold value for the drift detection.

    """

    warm_start: int = 500
    warning_threshold: float = 2.0
    drift_threshold: float = 3.0


class ParamsADWIN(OpBaseModel):
    """
    Parameters for the ADWIN (Adaptive Windowing) algorithm.

    Attributes:
        delta (float): The delta value for the drift detection.
        clock (int): The clock value for the drift detection.
        max_buckets (int): The maximum number of buckets to keep for the drift detection.
        min_window_length (int): The minimum length of the window for the drift detection.
        grace_period (int): The grace period value for the drift detection.

    """

    delta: float = 0.002
    clock: int = 32
    max_buckets: int = 5
    min_window_length: int = 5
    grace_period: int = 5


@register_op
class ConceptDrift(ColumnOp):
    """
    Operator for detecting concept drift using the DDM (Drift Detection Method) or ADWIN (Adaptive Windowing) algorithm.

    Attributes:
        algorithm (Literal["DDM", "ADWIN"]): The algorithm to use for concept drift detection.
        params (Union[ParamsDDM, ParamsADWIN]): The parameters for the selected algorithm.
        col_in_measure (str): The name of the column in the input data representing the metric to measure concept drift.

    Raises:
        ValueError: If the specified algorithm does not match the type of the parameters.

    Example:
        ```py
        import polars as pl
        from uptrain.operators import ParamsDDM, ConceptDrift

        # Create an instance of the ParamsDDM class with the parameters

        params_ddm = ParamsDDM(
                        warm_start=500,
                        warn_threshold=2.0,
                        alarm_threshold=3.0
                    )

        # Create an instance of the ConceptDrift operator
        op = ConceptDrift(
                algorithm="DDM",
                params=params_ddm,
                col_in_measure="metric"
            )

        # Set up the operator
        op.setup()

        # Run the operator on the input data
        input_data = pl.DataFrame(...)
        output = op.run(input_data)["extra"]

        # Check the detected concept drift information
        if output["alert_info"] is not None:
            print("Counter:", output["alert_info"]["counter"])
        ```

    Output:
        ```
        INFO     | uptrain.operators.drift:run:181 - Drift detected using DDM!
        Counter: 129466
        ```
    """

    algorithm: t.Literal["DDM", "ADWIN"]
    params: t.Union[ParamsDDM, ParamsADWIN]
    col_in_measure: str = "metric"

    @root_validator
    def _check_params(cls, values):
        """
        Check if the parameters are valid for the specified algorithm.

        """
        algo = values["algorithm"]
        params = values["params"]
        if algo == "DDM" and not isinstance(params, ParamsDDM):
            raise ValueError(
                f"Expected params to be of type {ParamsDDM} for algorithm - DDM"
            )
        elif algo == "ADWIN" and not isinstance(params, ParamsADWIN):
            raise ValueError(
                f"Expected params to be of type {ParamsADWIN} for algorithm - ADWIN"
            )
        return values

    def setup(self, settings: Settings):
        if self.algorithm == "DDM":
            self._algo_obj = drift.binary.DDM(**self.params.dict())  # type: ignore
        elif self.algorithm == "ADWIN":
            self._algo_obj = drift.ADWIN(**self.params.dict())  # type: ignore
        self._counter = 0
        self._avg_accuracy = 0.0
        self._cuml_accuracy = 0.0
        self._alert_info = None
        return self

    def run(self, data: pl.DataFrame) -> TYPE_TABLE_OUTPUT:
        ser = data.get_column(self.col_in_measure)

        for val in ser:
            self._algo_obj.update(val)
            if self._algo_obj.drift_detected and self._alert_info is None:
                msg = f"Drift detected using {self.algorithm}!"
                self._alert_info = {"counter": self._counter, "msg": msg}
                logger.info(msg)

            self._counter += 1
            self._cuml_accuracy += val

        self._avg_accuracy = (
            self._cuml_accuracy / self._counter if self._counter > 0 else 0.0
        )
        return {
            "output": None,
            "extra": {
                "counter": self._counter,
                "avg_accuracy": self._avg_accuracy,
                "alert_info": self._alert_info,
            },
        }



import os
from datetime import datetime, timedelta
import time
import pandas as pd
from uptrain.v0.core.lib.helper_funcs import dependency_required

try:
    from google.cloud import bigquery
except:
    bigquery = None

try:
    import psycopg2
except:
    psycopg2 = None


@dependency_required(bigquery, "'google-cloud-bigquery[pandas]'")
class BigQueryReader:
    def __init__(self, args) -> None:
        self.sql_query = args.sql_query
        self.client = bigquery.Client()
        self.sql_variables = args.sql_variables_dictn

    def run(self, time_diff):
        run_dict = {}
        for var, val in self.sql_variables.items():
            run_dict.update({
                var: str(val + timedelta(seconds=time_diff))
            })

        query = self.sql_query.format(**run_dict)
        print(f"Executing SQL query: {query} \n at UTC time: {str(datetime.utcnow())}")
        query_job = self.client.query(query)
        df = query_job.to_dataframe()
        print(f"Got {len(df)} rows")
        return df


@dependency_required(psycopg2, "psycopg2")
class PostgreSQLReader:
    def __init__(self, args) -> None:
        self.sql_query = args.sql_query
        self.conn = psycopg2.connect(
            database=args.database,
            user=os.environ['POSTGRES_USERNAME'],
            password=os.environ['POSTGRES_PASSWORD'])
        self.sql_variables = args.sql_variables_dictn

    def run(self, time_diff):
        run_dict = {}
        for var, val in self.sql_variables.items():
            run_dict.update({
                var: str(val + timedelta(seconds=time_diff))
            })

        query = self.sql_query.format(**run_dict)
        print(f"Executing SQL query: {query} \n at UTC time: {str(datetime.utcnow())}")
        df = pd.read_sql(query, self.conn)
        print(f"Got {len(df)} rows")
        return df


class RunTimeManager:
    def __init__(self, args, fw) -> None:
        self.framework = fw
        self.frequency_in_seconds = args.frequency_in_seconds
        self.num_backlog = args.num_backlog
        self.reader = self._resolve(args)
        self.time_diff = 0


    def run_endlessly(self):
        latest_run_time = datetime.utcnow() + timedelta(seconds=-self.frequency_in_seconds * self.num_backlog)
        while(1):
            if datetime.utcnow() > (latest_run_time + timedelta(seconds=self.frequency_in_seconds)):
                latest_run_time = latest_run_time + timedelta(seconds=self.frequency_in_seconds)
                data = self.reader.run(self.time_diff)
                if len(data) > 0:
                    self.framework.log(data)
                self.time_diff += self.frequency_in_seconds
            else:
                time.sleep(self.frequency_in_seconds * 0.1)


    def _resolve(self, args):
        if args.type == 'bigquery':
            return BigQueryReader(args)
        elif args.type == 'postgresql':
            return PostgreSQLReader(args)
        else:
            raise Exception(f"{args.type} is not supported!")


import numpy as np
import pytest
import uptrain.v0 as v0


@pytest.fixture(scope="module")
def random_state():
    return np.random.RandomState(seed=1337)


@pytest.fixture(scope="module")
def dataset(random_state):
    """Returns a dataset with 4 types of data points."""
    n = 200
    type_count = 4
    p = random_state.randn(type_count)
    p /= np.sum(p)
    labels = random_state.choice(np.arange(type_count), size=n, p=p)
    data = []
    for label in labels:
        data.append(random_state.normal(np.full(n, label), 1, n))
    return data, labels


@pytest.fixture(scope="module")
def umap_cfg():
    """Returns the parameters for UMAP."""
    return {
        "type": v0.Visual.UMAP,
        "measurable_args": {
            "type": v0.MeasurableType.INPUT_FEATURE,
            "feature_name": "data",
        },
        "label_args": [
            {
                "type": v0.MeasurableType.INPUT_FEATURE,
                "feature_name": "labels",
            }
        ],
        "min_dist": 0.01,
        "n_neighbors": 20,
        "metric": "euclidean",
        "update_freq": 1,
        "clustering_algorithm": v0.ClusteringAlgorithm.HDBSCAN,
    }


@pytest.fixture(scope="module")
def tsne_cfg():
    """Returns the parameters for t-SNE."""
    return {
        "type": v0.Visual.TSNE,
        "measurable_args": {
            "type": v0.MeasurableType.INPUT_FEATURE,
            "feature_name": "data",
        },
        "label_args": [
            {
                "type": v0.MeasurableType.INPUT_FEATURE,
                "feature_name": "labels",
            }
        ],
        "update_freq": 1,
        "perplexity": 10,
        "clustering_algorithm": v0.ClusteringAlgorithm.DBSCAN,
        "clustering_args": {
            "eps": 0.5,
            "min_samples": 5,
            "metric": "euclidean",
            "algorithm": "auto",
            "leaf_size": 30,
            "p": None,
            "n_jobs": None,
        },
    }


def test_visuals_dimensionality_reduction(dataset, umap_cfg, tsne_cfg):
    """Test Visuals Dimensionality Reduction.
    
    The test creates a UMAP and t-SNE visualisation of the data points in 2D
    and 3D. The test passes if the clusters are clearly visible in the visualisations.
    """
    data, labels = dataset

    cfg = {
        "checks": [
            {**umap_cfg, "dim": "3D", "dashboard_name": "umap_3d"},
            {**tsne_cfg, "dim": "3D", "dashboard_name": "tsne_3d"},
            # {**umap_cfg, "dim": "2D", "dashboard_name": "umap_2d"},
            # {**tsne_cfg, "dim": "2D", "dashboard_name": "tsne_2d"},
        ],
        "logging_args": {
            "log_folder": "uptrain_logs_dimensionality_reduction_1",
            "st_logging": True,
        },
    }

    # Set up the framework using the configurations
    framework = v0.Framework(cfg)

    # Log the inputs
    framework.log(inputs={"data": data, "labels": labels})


def test_visuals_dimensionality_reduction_new_logging(dataset, umap_cfg, tsne_cfg):
    data, labels = dataset

    cfg = {
        "checks": [
            {**umap_cfg, "dim": "3D", "dashboard_name": "umap_3d"},
            {**tsne_cfg, "dim": "3D", "dashboard_name": "tsne_3d"},
        ],
        "logging_args": {
            "log_folder": "uptrain_logs_dimensionality_reduction_2",
            "use_new_handler": True,
            "run_background_streamlit": False
        },
    }

    # Set up the framework using the configurations
    framework = v0.Framework(cfg)

    # Log the inputs
    _ = framework.log(inputs={"data": data, "labels": labels})




from __future__ import annotations
import os
import typing as t

from loguru import logger
import polars as pl
from uptrain.framework.evals import Evals, ParametricEval
from uptrain.framework.evalllm import EvalLLM
from uptrain.framework.remote import APIClient
from uptrain.framework.remote import DataSchema


from llama_index.indices.query.base import BaseQueryEngine


__all__ = [
    "EvalLlamaIndex"
]

class EvalLlamaIndex:
    query_engine: BaseQueryEngine

    def __init__(self, query_engine: BaseQueryEngine) -> None: 
        if not isinstance(query_engine, BaseQueryEngine):
            raise Exception("Please provide Query Engine for the evaluation")
        self.query_engine = query_engine

    def evaluate(
        self,
        client : t.Union[EvalLLM, APIClient],
        data: t.Union[list[dict], pl.DataFrame],
        checks: list[t.Union[str, Evals, ParametricEval]],
        project_name: str = None,
        schema: t.Union[DataSchema, dict[str, str], None] = None,
        metadata: t.Optional[dict[str, str]] = None,
    ):
        try:
            from llama_index.async_utils import run_async_tasks
        except ImportError:
            raise ImportError(
                "llama_index must be installed to use this function. "
                "Install it with `pip install llama_index`."
            )
        """
        Run evaluation of llama_index QueryEngine with different Evals

        NOTE: This api doesn't log any data.
        Args:
            project_name: Name of the project to evaluate on.
            client: EvalLLM or APIClient object used for the evaluation using user's openai keys or Uptrain API key.
            data: Data to evaluate on. Either a Polars DataFrame or a list of dicts.
            checks: List of checks to evaluate on.
            schema: Schema of the data. Only required if the data attributes aren't typical (question, response, context).
            metadata: Attributes to attach to this dataset. Useful for filtering and grouping in the UI.
        Returns:
            results: List of dictionaries with each data point and corresponding evaluation results.
        """
        
        if isinstance(data, pl.DataFrame):
            data = data.to_dicts()
        import nest_asyncio
        nest_asyncio.apply()
        
        if schema is None:
            schema = DataSchema()
        elif isinstance(schema, dict):
            schema = DataSchema(**schema)
        
        responses = run_async_tasks([self.query_engine.aquery(data[i][schema.question]) for i in range(len(data))])

        for index, r in enumerate(responses):
            data[index][schema.response] = r.response
            data[index][schema.context] = "\n".join([c.node.get_content() for c in r.source_nodes])

        if isinstance(client, EvalLLM):
            results = client.evaluate(
                data = data,
                checks = checks,
                schema = schema,
                metadata = metadata
            )
        elif isinstance(client, APIClient):
            results = client.log_and_evaluate(
                project_name = project_name,
                data = data,
                checks = checks,
                schema = schema,
                metadata = metadata
            )
        return results


"""IO operators to read from Excel files"""

from __future__ import annotations
import typing as t

import polars as pl
import deltalake as dl

if t.TYPE_CHECKING:
    from uptrain.framework import Settings
from uptrain.operators.base import *
from uptrain.utilities import lazy_load_dep

xlsx2csv = lazy_load_dep("xlsx2csv", "xlsx2csv")


@register_op
class ExcelReader(TransformOp):
    """Reads an excel file.

    Attributes:
        fpath (str): Path to the xlsx file.
        batch_size (Optional[int]): Number of rows to read at a time. Defaults to None, which reads the entire file.

    """

    fpath: str
    batch_size: t.Optional[int] = None

    def setup(self, settings: Settings):
        return self

    def run(self) -> TYPE_TABLE_OUTPUT:
        return {"output": pl.read_excel(self.fpath)}



import uptrain.v0 as v0
from test_helpers import KpsDataset, get_data_from_remote
import torch
from test_helpers import body_length_signal, train_model_torch, BinaryClassification


def test_data_integrity():
    get_data_from_remote()

    inference_batch_size = 16

    # Training data
    training_file = "data/training_data.json"

    # A testing dataset to evaluate model performance
    golden_testing_file = "data/golden_testing_data.json"

    # The data-points which the models sees in production
    real_world_test_cases = "data/real_world_testing_data.json"

    # To annotate the collected data points, we extract the ground truth from a master annotation file
    # (we can also import from any other annotation pipelines such as an annotation job on Mechanical turk).
    annotation_args = {"master_file": "data/master_annotation_data.json"}

    train_model_torch(training_file, "version_0")

    cfg = {
        "checks": [
            {
                "type": v0.Monitor.DATA_INTEGRITY,
                "measurable_args": {
                    "type": v0.MeasurableType.INPUT_FEATURE,
                    "feature_name": "kps",
                },
                "integrity_type": "non_null",
            },
            {
                "type": v0.Monitor.DATA_INTEGRITY,
                "measurable_args": {
                    "type": v0.MeasurableType.CUSTOM,
                    "signal_formulae": v0.Signal("body_length", body_length_signal),
                },
                "integrity_type": "greater_than",
                "threshold": 50,
            },
        ],
        "retraining_folder": "uptrain_smart_data_data_integrity",
    }

    framework_torch = v0.Framework(cfg)

    model_dir = "trained_models_torch/"
    model_save_name = "version_0"
    real_world_dataset = KpsDataset(
        real_world_test_cases, batch_size=inference_batch_size, is_test=True
    )
    model = BinaryClassification()
    model.load_state_dict(torch.load(model_dir + model_save_name))
    model.eval()

    for i, elem in enumerate(real_world_dataset):
        # Do model prediction
        inputs = {"kps": elem[0]["kps"], "id": elem[0]["id"]}
        x_test = torch.tensor(inputs["kps"]).type(torch.float)
        test_logits = model(x_test).squeeze()
        preds = torch.round(torch.sigmoid(test_logits)).detach().numpy()

        # Log model inputs and outputs to the uptrain Framework to monitor data integrity
        idens = framework_torch.log(inputs=inputs, outputs=preds)



__all__ = [
    "EvalLLM",
    "APIClient",
    "Evals",
    "ConversationSatisfaction",
    "CritiqueTone",
    "GuidelineAdherence",
    "ResponseMatching",
    "Settings",
    "EvalLlamaIndex"
]

from .framework.base import Settings
from .framework.remote import APIClient
from .framework.evals import Evals, CritiqueTone, GuidelineAdherence, ResponseMatching, ConversationSatisfaction
from .framework.evalllm import EvalLLM

from .integrations.llama_index import EvalLlamaIndex



"""Utility routines to cache data when computing metrics
"""

from __future__ import annotations
from abc import abstractmethod
from typing import Any, Type, Protocol

import numpy as np


# -----------------------------------------------------------
# State cache
# -----------------------------------------------------------


class StateCache(Protocol):
    """Container to store intermediate state when computing aggregate statistics."""

    def __init__(self, fw: Any, columns: dict[str, Type]):
        pass

    @abstractmethod
    def upsert_ids_n_col_values(
        self, ids: np.ndarray, col_values: dict[str, np.ndarray]
    ):
        pass

    @abstractmethod
    def fetch_col_values_for_ids(
        self, ids: np.ndarray, col_names: list[str]
    ) -> list[dict]:
        pass


class DictStateCache:
    """A state cache that uses python dictionaries to store intermediate state."""

    def __init__(self, fw: Any, columns: dict[str, Type]) -> None:
        self.state_dicts = {col: {} for col in columns}

    def __del__(self) -> None:
        for container in self.state_dicts.values():
            container.clear()
        del self.state_dicts

    def upsert_ids_n_col_values(
        self, ids: np.ndarray, col_values: dict[str, np.ndarray]
    ) -> None:
        for col, values in col_values.items():
            assert len(ids) == len(values)
            container = self.state_dicts[col]
            container.update({id: value for id, value in zip(ids, values)})

    def fetch_col_values_for_ids(
        self, ids: np.ndarray, col_names: list[str]
    ) -> list[dict]:
        output = []
        for col in col_names:
            container = self.state_dicts[col]
            output.append({id: container[id] for id in ids if id in container})
        return output


def make_cache_container(fw: Any, columns: dict[str, Type]) -> StateCache:
    """Create a state cache container for the given columns."""
    if fw.config_obj.running_ee:
        try:
            from uptrain.v0.ee.lib.cache import DuckDBStateCache
        except ImportError as exc:
            print(
                "Error importing the uptrain-ee subpackage, Uptrain enterprise Edition not installed."
            )
            raise exc

        return DuckDBStateCache(fw, columns)
    else:
        return DictStateCache(fw, columns)



import json
import numpy as np
import cv2
import os
import pandas as pd
import imgaug as ia
import imgaug.augmenters as iaa
from imgaug.augmentables import Keypoint
from sklearn.preprocessing import Normalizer


class KpsDataset:
    def __init__(
        self,
        file_name,
        n_classes=2,
        batch_size=16,
        shuffle=False,
        augmentations=False,
        is_test=False,
        normalization=False,
    ):
        data = read_json(file_name)
        self.data = np.array(data)
        self.n_classes = n_classes
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.augmentations = augmentations
        self.is_test = is_test
        self.normalization = normalization
        np.random.seed(10)
        ia.seed(10)
        self.on_epoch_end()

    def load_x_y_from_data(self, data=None):
        if data is None:
            data = self.data
        kps = [x["kps"] for x in data]
        gts = [x["gt"] for x in data]
        ids = [x["id"] for x in data]
        if self.normalization:
            norm = Normalizer()
            kps = norm.fit_transform(kps)
        return (kps, gts, ids)

    def on_epoch_end(self):
        self.indexes = np.arange(len(self.data))
        if self.shuffle:
            np.random.shuffle(self.indexes)

    def __len__(self):
        return int(np.floor(len(self.indexes) / self.batch_size))

    def __getitem__(self, index):
        indexes = self.indexes[index * self.batch_size : (index + 1) * self.batch_size]
        this_data = [self.data[i] for i in indexes]
        batch_kps, batch_gts, batch_ids = self.load_x_y_from_data(this_data)

        for idx in range(len(batch_kps)):
            raw_kps = np.reshape(np.array(batch_kps[idx]), (17, 2))
            kps = []
            for ijdx in range(17):
                kps.append(Keypoint(x=raw_kps[ijdx][0], y=raw_kps[ijdx][1]))
            image = np.zeros((2, 2, 3))

            if self.augmentations:
                seq = iaa.Sequential(
                    [
                        iaa.Affine(
                            rotate=(-10, 10),
                            scale=(0.75, 1.25),
                            translate_percent=(-0.25, 0.25),
                            shear=(-20, 20),
                        )
                    ]
                )
                image_aug, kps_aug = seq(image=image, keypoints=kps)
                new_kps = []
                for ijdx in range(len(kps_aug)):
                    new_kps.extend([kps_aug[ijdx].x, kps_aug[ijdx].y])
            else:
                new_kps = np.reshape(raw_kps, 34)
            batch_kps[idx] = np.array(new_kps)

        if self.is_test:
            sample = (
                {"kps": np.array(batch_kps), "id": np.array(batch_ids)},
                np.array(batch_gts),
            )
        else:
            sample = (np.array(batch_kps), np.array(batch_gts))

        return sample

    def __iter__(self):
        """Create a generator that iterate over the Sequence."""
        for item in (self[i] for i in range(len(self))):
            yield item


def read_json(file_name, dataframe=False):
    with open(file_name) as f:
        data = json.load(f)
    if dataframe:
        data = pd.DataFrame(data)
        data = data.drop(["frame_idx"], axis=1)
        keys = [
            "Nose_X",
            "Nose_Y",
            "Left_Eye_X",
            "Left_Eye_Y",
            "Right_Eye_X",
            "Right_Eye_Y",
            "Left_Ear_X",
            "Left_Ear_Y",
            "Right_Ear_X",
            "Right_Ear_Y",
            "Left_Shoulder_X",
            "Left_Shoulder_Y",
            "Right_Shoulder_X",
            "Right_Shoulder_Y",
            "Left_Elbow_X",
            "Left_Elbow_Y",
            "Right_Elbow_X",
            "Right_Elbow_Y",
            "Left_Wrist_X",
            "Left_Wrist_Y",
            "Right_Wrist_X",
            "Right_Wrist_Y",
            "Left_Hip_X",
            "Left_Hip_Y",
            "Right_Hip_X",
            "Right_Hip_Y",
            "Left_Knee_X",
            "Left_Knee_Y",
            "Right_Knee_X",
            "Right_Knee_Y",
            "Left_Ankle_X",
            "Left_Ankle_Y",
            "Right_Ankle_X",
            "Right_Ankle_Y",
        ]
        for idx in range(len(keys)):
            values = [x[idx] for x in list(data["kps"])]
            data[keys[idx]] = values
        data = data.drop(["kps"], axis=1)
    return data


def write_json(file_name, data):
    with open(file_name, "w") as f:
        json.dump(data, f)


def plot_all_cluster(all_clusters, num_labels):
    for idx in range(len(all_clusters)):
        frame = plot_kps_as_image(
            all_clusters[idx], int(100 * num_labels[idx] / sum(num_labels))
        )
        cv2.imwrite(str(idx) + ".png", frame)

    # TODO: Plot images so that size of the image is proportional to the count of that cluster
    plot_cluster_as_image(len(all_clusters))


def plot_cluster_as_image(num_clusters):
    for idx in range(0, int(num_clusters / 5)):
        for jdx in range(0, 5):
            if jdx == 0:
                frame = cv2.imread(str(idx * 5 + jdx) + ".png")
            else:
                this_frame = cv2.imread(str(idx * 5 + jdx) + ".png")
                frame = cv2.hconcat([frame, this_frame])
            os.remove(str(idx * 5 + jdx) + ".png")
        if idx == 0:
            full_frame = frame
        else:
            full_frame = cv2.vconcat([full_frame, frame])
    file_name = "num_clusters_" + str(num_clusters)
    cv2.imwrite(file_name + ".png", full_frame)


def plot_kps_as_image(kps, prob):
    frame = np.zeros((256, 256, 3))
    max_val = max(kps)
    kps = kps * 200 / max_val
    kps = np.array(kps).reshape((17, 2))

    h, w = frame.shape[0], frame.shape[1]
    color = [256, 256, 256]

    for iter_keypoints in range(17):
        x_raw_cord = int(np.ceil(kps[iter_keypoints][1]))
        y_raw_cord = int(np.ceil(kps[iter_keypoints][0]))
        radius = 3
        frame[
            min(max(0, x_raw_cord - radius), h) : max(0, min(x_raw_cord + radius, h)),
            min(max(0, y_raw_cord - radius), w) : max(0, min(y_raw_cord + radius, w)),
        ] = color

    frame = cv2.UMat(frame)

    skeleton_connections = [
        [[16, 14], [14, 12]],
        [[17, 15], [15, 13]],
        [[12, 13], [6, 12], [7, 13], [6, 7]],
        [[6, 8], [8, 10]],
        [[7, 9], [9, 11]],
        [[2, 3], [1, 2], [1, 3], [2, 4], [3, 5]],
    ]
    keypoints = kps
    thickness = 1

    for skeleton_connection in skeleton_connections:
        for iter_line in range(len(skeleton_connection)):
            start_point = (
                int(np.ceil(keypoints[skeleton_connection[iter_line][0] - 1][0])),
                int(np.ceil(keypoints[skeleton_connection[iter_line][0] - 1][1])),
            )
            end_point = (
                int(np.ceil(keypoints[skeleton_connection[iter_line][1] - 1][0])),
                int(np.ceil(keypoints[skeleton_connection[iter_line][1] - 1][1])),
            )
            frame = cv2.line(frame, start_point, end_point, color, thickness)

    cv2.putText(
        frame,
        "Prob: " + str(prob),
        (5, 15),
        cv2.FONT_HERSHEY_SIMPLEX,
        0.5,
        (255, 255, 255),
        1,
        2,
    )
    return frame


def plot_all_cluster(all_clusters, num_labels, plot_save_name=""):
    for idx in range(len(all_clusters)):
        frame = plot_kps_as_image(all_clusters[idx], int(num_labels[idx]))
        cv2.imwrite(str(idx) + ".png", frame)

    # TODO: Plot images so that size of the image is proportional to the count of that cluster
    plot_cluster_as_image(len(all_clusters), plot_save_name=plot_save_name)


def plot_cluster_as_image(num_clusters, plot_save_name=""):
    for idx in range(0, int(np.ceil(num_clusters / 5))):
        for jdx in range(0, 5):
            if idx * 5 + jdx >= num_clusters:
                frame = cv2.hconcat([frame, this_frame * 0])
                continue
            if jdx == 0:
                frame = cv2.imread(str(idx * 5 + jdx) + ".png")
            else:
                this_frame = cv2.imread(str(idx * 5 + jdx) + ".png")
                frame = cv2.hconcat([frame, this_frame])
            os.remove(str(idx * 5 + jdx) + ".png")
        if idx == 0:
            full_frame = frame
        else:
            full_frame = cv2.vconcat([full_frame, frame])
    if plot_save_name == "":
        plot_save_name = "num_clusters_" + str(num_clusters) + ".png"
    cv2.imwrite(plot_save_name, full_frame)


def plot_kps_as_image(kps, prob):
    frame = np.zeros((256, 256, 3))
    max_val = max(kps)
    kps = kps * 200 / max_val
    kps = np.array(kps).reshape((17, 2))

    h, w = frame.shape[0], frame.shape[1]
    color = [256, 256, 256]

    for iter_keypoints in range(17):
        x_raw_cord = int(np.ceil(kps[iter_keypoints][1]))
        y_raw_cord = int(np.ceil(kps[iter_keypoints][0]))
        radius = 3
        frame[
            min(max(0, x_raw_cord - radius), h) : max(0, min(x_raw_cord + radius, h)),
            min(max(0, y_raw_cord - radius), w) : max(0, min(y_raw_cord + radius, w)),
        ] = color

    frame = cv2.UMat(frame)

    skeleton_connections = [
        [[16, 14], [14, 12]],
        [[17, 15], [15, 13]],
        [[12, 13], [6, 12], [7, 13], [6, 7]],
        [[6, 8], [8, 10]],
        [[7, 9], [9, 11]],
        [[2, 3], [1, 2], [1, 3], [2, 4], [3, 5]],
    ]
    keypoints = kps
    thickness = 1

    for skeleton_connection in skeleton_connections:
        for iter_line in range(len(skeleton_connection)):
            start_point = (
                int(np.ceil(keypoints[skeleton_connection[iter_line][0] - 1][0])),
                int(np.ceil(keypoints[skeleton_connection[iter_line][0] - 1][1])),
            )
            end_point = (
                int(np.ceil(keypoints[skeleton_connection[iter_line][1] - 1][0])),
                int(np.ceil(keypoints[skeleton_connection[iter_line][1] - 1][1])),
            )
            frame = cv2.line(frame, start_point, end_point, color, thickness)

    cv2.putText(
        frame,
        "Support: " + str(prob),
        (5, 15),
        cv2.FONT_HERSHEY_SIMPLEX,
        0.5,
        (255, 255, 255),
        1,
        2,
    )
    return frame



from typing import Any
import numpy as np

from uptrain.v0.core.classes.measurables import Measurable


class FeatureConcatMeasurable(Measurable):
    """Class that returns the input feature corresponding to the feature name."""

    def __init__(self, framework, feat_name_list) -> None:
        super().__init__(framework)
        self.feat_name_list = feat_name_list

    def _compute(self, inputs=None, outputs=None, gts=None, extra=None) -> Any:
        vals = [inputs[x] for x in self.feat_name_list]
        vals = np.transpose(vals)
        return vals

    def col_name(self) -> str:
        name = 'concat_' + '_'.join(self.feat_name_list)
        return name

    # TODO: Decommission and find a generic way
    def extract_val_from_training_data(self, x):
        vals = [x[y] for y in self.feat_name_list]
        return vals



"""A pytest module to test that the SQLite datastore works as expected.
"""

import os
import sqlite3
import tempfile

import pytest
import numpy as np

from uptrain.v0.core.lib.datastores import SqliteStore


@pytest.fixture
def sqlite_store():
    """Fixture to create a temporary SQLite datastore."""
    with tempfile.TemporaryDirectory() as tmpdir:
        path = os.path.join(tmpdir, "test")
        yield SqliteStore(path)


def test_table_created(sqlite_store: SqliteStore):
    """Test that the SQLite datastore works as expected."""
    datapoint = {"id": 1, "val": 2.0, "name": "test"}
    sqlite_store.log(datapoint)

    with sqlite3.connect(sqlite_store.path_db) as conn:
        cursor = conn.cursor()
        cursor.execute("SELECT id, val, name FROM logs")
        assert cursor.fetchone() == (1, 2.0, "test")

        res = cursor.execute("SELECT NAME FROM PRAGMA_TABLE_INFO('logs')")
        assert res.fetchall() == [("id",), ("val",), ("name",), ("internal_ts",)]


def test_batch_logging(sqlite_store: SqliteStore):
    """Test logging a batch of data at once works as expected."""
    alphabet = list("abcdefghijklmnopqrstuvwxyz")
    datapoints = {
        "id": np.arange(1000),  # int
        "feat_1": np.random.randn(1000),  # float
        "feat_2": ["".join(x) for x in np.random.choice(alphabet, (1000, 5))],  # str
        "feat_3": np.random.choice([True, False], 1000),  # bool
        "feat_4": np.arange(6000).reshape(1000, 6),  # np.ndarray
    }
    sqlite_store.log_many(datapoints)
    sqlite_store.log_many(datapoints)

    with sqlite3.connect(sqlite_store.path_db) as conn:
        cursor = conn.cursor()
        cursor.execute("SELECT id, feat_1, feat_2, feat_3, feat_4 FROM logs")
        res = cursor.fetchall()
        assert len(res) == 2000



import numpy as np

from uptrain.v0.core.classes.measurables import Measurable


class AccuracyMeasurable(Measurable):
    """Class that computes the accuracy from predictions and ground truths.

    Returns a list of values where a value is:
    - True, if prediction equals ground truth
    - False, if prediction does not equal ground truth
    """

    def __init__(self, framework) -> None:
        super().__init__(framework)

    def _compute(self, inputs=None, outputs=None, gts=None, extra=None) -> np.ndarray:
        gts = np.array(gts)
        outputs = np.reshape(np.array(outputs), gts.shape)
        return np.equal(outputs, gts)

    def col_name(self) -> str:
        return "Accuracy"


class MAEMeasurable(Measurable):
    """Class that computes the Mean Absolute Error from predictions and ground truths.

    Returns a list of values where each value is the absolute value of the
    difference between corresponding prediction and ground truth.
    """

    def __init__(self, framework) -> None:
        super().__init__(framework)

    def _compute(self, inputs=None, outputs=None, gts=None, extra=None) -> np.ndarray:
        gts = np.array(gts)
        outputs = np.reshape(np.array(outputs), gts.shape)
        return np.squeeze(np.abs(outputs - gts))

    def col_name(self) -> str:
        return "MAE"


class MAPEMeasurable(Measurable):
    """Class that computes the Mean Absolute Percentage Error from predictions and ground truths.

    Returns a list of values where each value is the absolute percentage
    difference between corresponding prediction and ground truth,
    multiplied by 100 to express the result as a percentage.
    """
    
    def __init__(self, framework) -> None:
        super().__init__(framework)

    def _compute(self, inputs=None, outputs=None, gts=None, extra=None) -> np.ndarray:
        gts = np.array(gts)
        outputs = np.reshape(np.array(outputs), gts.shape)
        return 100 * np.squeeze(np.abs(np.divide(outputs - gts, gts)))

    def col_name(self) -> str:
        return "MAPE"



from __future__ import annotations
import os
import sys
import typing as t
import uuid
import itertools
import numpy as np

import evals
import evals.base
import evals.record
import evals.registry
from loguru import logger
import polars as pl

if t.TYPE_CHECKING:
    from uptrain.framework import Settings
from uptrain.operators.base import *
from uptrain.utilities import to_py_types

UPTRAIN_BASE_DIR = os.path.dirname(os.path.abspath(__file__))

# -----------------------------------------------------------
# General purpose OpenAI eval operator. It can take any eval
# specified in the opanai-evals format and run it.
# -----------------------------------------------------------


class UptrainEvalRecorder(evals.record.RecorderBase):
    """Subclass the default OpenAI eval recorder so we don't need to write
    to temporary files.
    """

    _run_data: dict

    def __init__(self, run_spec: "evals.base.RunSpec"):
        super().__init__(run_spec)
        self._run_data = to_py_types(run_spec)

    def get_list_events(self, _type: t.Optional[str] = None) -> list[dict]:
        events = [to_py_types(event) for event in self._events]
        if _type is None:
            return events
        else:
            return [evt for evt in events if evt["type"] == _type]

    def get_run_data(self) -> dict:
        return self._run_data


@register_op
class OpenaiEval(ColumnOp):
    """
    Operator for running OpenAI evals.

    Attributes:
        bundle_path (str): Path to the bundle containing eval resources.
        completion_name (str): Name of the completion function to use.
        eval_name (str): Name of the eval to run.

    """

    bundle_path: str
    completion_name: str
    eval_name: str

    def setup(self, settings: Settings):
        import openai

        openai.api_key = settings.check_and_get("openai_api_key")
        return self

    def run(self, data: pl.DataFrame) -> TYPE_TABLE_OUTPUT:
        registry = evals.registry.Registry()
        registry_path = os.path.join(self.bundle_path, "custom_registry")
        registry.add_registry_paths([registry_path])

        eval_name = self.eval_name
        eval_spec = registry.get_eval(eval_name)
        assert (
            eval_spec is not None
        ), f"Eval {eval_name} not found. Available: {list(sorted(registry._evals.keys()))}"

        eval_name = eval_spec.key
        assert eval_name is not None

        # NOTE: create a temporary file with the samples if we are overriding the dataset
        path_samples_file = f"/tmp/{uuid.uuid4()}.jsonl"
        data.write_ndjson(path_samples_file)
        eval_spec.args["samples_jsonl"] = path_samples_file  # type: ignore

        # NOTE: add `custom_fns` to the python path"
        if self.bundle_path not in sys.path:
            sys.path.append(self.bundle_path)
        completion_fns = [self.completion_name]
        completion_fn_instances = [
            registry.make_completion_fn(url) for url in completion_fns
        ]

        run_config = {
            "completion_fns": completion_fns,
            "eval_spec": eval_spec,
            "seed": 42,
            "max_samples": None,
            "command": "",
            "initial_settings": {"visible": True},
        }
        run_spec = evals.base.RunSpec(
            completion_fns=completion_fns,
            eval_name=eval_name,
            base_eval=eval_name.split(".")[0],
            split=eval_name.split(".")[1],
            run_config=run_config,
            created_by="uptrain",
        )
        recorder = UptrainEvalRecorder(run_spec=run_spec)

        eval_class = registry.get_class(eval_spec)
        extra_eval_params = {}
        eval = eval_class(
            completion_fns=completion_fn_instances,
            seed=42,
            name=eval_name,
            registry=registry,
            **extra_eval_params,
        )
        final_report = eval.run(recorder)

        if path_samples_file is not None:
            os.remove(path_samples_file)

        extra = {
            "all_events": recorder.get_list_events(),
            "run_data": recorder.get_run_data(),
            "final_report": to_py_types(final_report),
        }
        unique_types = set(x["type"] for x in recorder.get_list_events())
        for typ in unique_types:
            extra[typ] = pl.from_dicts(
                [
                    x["data"]
                    for x in sorted(
                        recorder.get_list_events(typ),
                        key=lambda x: int(x["sample_id"].split(".")[-1]),
                    )
                ]
            )

        return {"output": None, "extra": extra}


# -----------------------------------------------------------
# Prompt eval operator
# -----------------------------------------------------------


class PromptEval(TransformOp):
    """
    Operator for running prompt-based evaluations.

    Attributes:
        prompt_template (str): Template for the prompt string.
        prompt_variables (list[str]): List of variables to substitute in the prompt template.
        gt_variables (list[str]): List of ground truth variables.
        model_name (str): Name of the model to use for evaluation.
        col_out_prompt (str, optional): Output column name for prompts. Defaults to "prompt".
        col_out_response (str, optional): Output column name for responses. Defaults to "response".
        _settings (Settings): The framework settings.

    """

    prompt_template: str
    prompt_variables: list[str]
    gt_variables: list[str]
    model_name: str
    col_out_prompt: str = "prompt"
    col_out_response: str = "response"
    _settings: Settings

    def setup(self, settings: Settings):
        self._settings = settings
        return self

    def _validate_data(self, data: pl.DataFrame) -> None:
        for col in self.prompt_variables:
            assert (
                col in data.columns
            ), f"Column for the prompt variable: {col} not found in input data."
        for col in self.gt_variables:
            assert (
                col in data.columns
            ), f"Column for the ground truth variable: {col} not found in input data."

    def _construct_prompts(self, data: pl.DataFrame) -> pl.DataFrame:
        prompts = [
            self.prompt_template.format(**{k: row[k] for k in self.prompt_variables})
            for row in data.rows(named=True)
        ]
        return pl.from_dict({"input": prompts})

    def run(self, data: pl.DataFrame) -> TYPE_TABLE_OUTPUT:
        self._validate_data(data)
        prompts = self._construct_prompts(data)

        eval_op = OpenaiEval(
            bundle_path=os.path.join(UPTRAIN_BASE_DIR, "openai_eval_custom"),
            completion_name=self.model_name,
            eval_name="model_run_all",
        )

        eval_op.setup(self._settings)
        oaieval_res = eval_op.run(prompts)
        assert "extra" in oaieval_res and "sampling" in oaieval_res["extra"]
        results = oaieval_res["extra"]["sampling"]

        return {
            "output": data.with_columns(
                [
                    pl.Series(results["prompt"]).alias(self.col_out_prompt),
                    pl.Series(
                        list(itertools.chain(*results["sampled"])),
                    ).alias(self.col_out_response),
                ]
            )
        }



"""
This module implements a simple client that can be used to schedule unit-tests/evaluations 
on the UpTrain server. 
"""

import typing as t

from loguru import logger
import httpx
import polars as pl
import pandas as pd
import pydantic

from uptrain.framework.checks import CheckSet, ExperimentArgs
from uptrain.framework.base import Settings
from uptrain.framework.evals import Evals, ParametricEval, CritiqueTone, GuidelineAdherence, ResponseMatching, ConversationSatisfaction


class DataSchema(pydantic.BaseModel):
    id_: str = "id"
    question: str = "question"
    response: str = "response"
    context: str = "context"
    ground_truth: str = "ground_truth"
    conversation: str = 'conversation'
    scenario: str = "scenario"
    objective: str = "objective"


def raise_or_return(response: httpx.Response):
    if not response.is_success:
        logger.error(response.text)
        response.raise_for_status()
    else:
        return response.json()


class APIClientWithoutAuth:
    base_url: str
    client: httpx.Client

    def __init__(self, settings: Settings = None) -> None:
        if settings is None:
            settings = Settings()

        server_url = settings.check_and_get("uptrain_server_url")
        self.base_url = server_url.rstrip("/") + "/api/open"
        self.client = httpx.Client(
            timeout=httpx.Timeout(500, connect=5),
        )

    def evaluate(
        self,
        data: list[dict],
        checks: list[t.Union[Evals, ParametricEval]],
        metadata: dict,
    ):
        """Run an evaluation on the UpTrain server (Doesn't require UpTrain API Key).
        """

        url = f"{self.base_url}/evaluate_no_auth"
        response_json = []
        try:
            response = self.client.post(
                url,
                json={
                    "data": data,
                    "checks": checks,
                    "metadata": metadata
                },
            )
            response_json = raise_or_return(response)
        except Exception as e:
            logger.error(f"Evaluation failed with error: {e}")
            raise e

        return response_json



class APIClient:
    base_url: str
    client: httpx.Client

    def __init__(self, settings: Settings = None, uptrain_api_key: str = None) -> None:
        if (uptrain_api_key is None) and (settings is None):
            raise Exception("Please provide UpTrain API Key")

        if settings is None:
            settings = Settings(uptrain_access_token=uptrain_api_key)

        server_url = settings.check_and_get("uptrain_server_url")
        api_key = settings.check_and_get("uptrain_access_token")
        self.settings = settings
        self.base_url = server_url.rstrip("/") + "/api/public"
        self.client = httpx.Client(
            headers={"uptrain-access-token": api_key},
            timeout=httpx.Timeout(1500, connect=5),
        )

    def check_auth(self):
        """Ping the server to check if the client is authenticated."""
        url = f"{self.base_url}/auth"
        try:
            response = self.client.get(url)
            return raise_or_return(response)
        except httpx.ConnectError as e:
            raise RuntimeError(
                f"Failed to connect to the Uptrain server at {self.base_url}"
            ) from e

    def add_dataset(self, name: str, fpath: str):
        url = f"{self.base_url}/dataset"
        with open(fpath, "rb") as file:
            files = {"data_file": (name, file, "application/octet-stream")}
            response = self.client.post(url, data={"name": name}, files=files)
        return raise_or_return(response)

    def get_dataset(self, name: str, version: t.Optional[int] = None):
        url = f"{self.base_url}/dataset"
        params: dict = {"name": name}
        if version is not None:
            params["version"] = version
        response = self.client.get(url, params=params)
        return raise_or_return(response)

    def list_datasets(self, skip: int = 0, limit: int = 100):
        url = f"{self.base_url}/datasets"
        params = {"skip": skip, "limit": limit}
        response = self.client.get(url, params=params)
        return raise_or_return(response)

    def download_dataset(
        self, name: str, fpath: str, version: t.Optional[int] = None
    ) -> None:
        """
        Download a dataset from the server.

        Args:
            name: name of the dataset to download
            fpath: path to save the dataset to
            version: version of the dataset to download. If None, the latest version is downloaded.
        """
        url = f"{self.base_url}/dataset/{name}/download"
        params: dict = {}
        if version is not None:
            params["version"] = version
        response = self.client.get(url, params=params)
        if not response.is_success:
            logger.error(response.text)
            response.raise_for_status()
        else:
            with open(fpath, "wb") as download_file:
                for chunk in response.iter_bytes():
                    download_file.write(chunk)

    def add_checkset(self, name: str, checkset: CheckSet, settings: Settings):
        url = f"{self.base_url}/checkset"
        response = self.client.post(
            url,
            json={"name": name, "config": checkset.dict(), "settings": settings.dict()},
        )
        return raise_or_return(response)

    def get_checkset(self, name: str, version: t.Optional[int] = None):
        url = f"{self.base_url}/checkset"
        params: dict = {"name": name}
        if version is not None:
            params["version"] = version
        response = self.client.get(url, params=params)
        return raise_or_return(response)

    def list_checksets(self, skip: int = 0, limit: int = 10):
        url = f"{self.base_url}/checksets"
        params = {"skip": skip, "limit": limit}
        response = self.client.get(url, params=params)
        return raise_or_return(response)

    def add_experiment(
        self,
        name: str,
        checkset: CheckSet,
        experiment_args: ExperimentArgs,
        settings: Settings,
    ):
        preprocessors = experiment_args._get_preprocessors()
        modified_checks = experiment_args._modify_checks(checkset.checks)
        modified_checkset = CheckSet(
            source=checkset.source, checks=modified_checks, preprocessors=preprocessors
        )
        url = f"{self.base_url}/checkset"
        response = self.client.post(
            url,
            json={
                "name": name,
                "config": modified_checkset.dict(),
                "settings": settings.dict(),
            },
        )
        return raise_or_return(response)

    def add_run(self, dataset: str, checkset: str):
        """Schedules an evaluation on the server. Specify the dataset and checkset to use.

        Args:
            dataset: name of the dataset to use
            checkset: name of the checkset to use

        Returns:
            run: information about the run along with a unique identifier.
        """
        url = f"{self.base_url}/run"
        response = self.client.post(
            url, json={"dataset": dataset, "checkset": checkset}
        )
        return raise_or_return(response)

    def get_run(self, run_id: str):
        """Get the status of a run.

        Args:
            run_id: unique identifier for the run.

        Returns:
            run: information about the run along with a unique identifier.
        """
        url = f"{self.base_url}/run/{run_id}"
        response = self.client.get(url)
        return raise_or_return(response)

    def get_run_results(self, run_id: str, check_name: str) -> list[dict]:
        """Get the results of a run.

        Args:
            run_id: unique identifier for the run.
            check_name: name of the check to get results for.

        Returns:
            run: information about the run along with a unique identifier.
        """
        url = f"{self.base_url}/run/{run_id}/results"
        params: dict = {"check_name": check_name}
        response = self.client.get(url, params=params)
        if not response.is_success:
            logger.error(response.text)
            response.raise_for_status()
        else:
            return pl.read_ndjson(response.content).to_dicts()

    def download_run_result(self, run_id: str, check_name: str, fpath: str) -> None:
        """Download the results of a run.

        Args:
            run_id: unique identifier for the run.
            check_name: name of the check to get results for.
            fpath: path to save the results to.
        """
        url = f"{self.base_url}/run/{run_id}/results"
        params: dict = {"check_name": check_name}
        with self.client.stream("GET", url, params=params) as response:
            if not response.is_success:
                logger.error(response.text)
                response.raise_for_status()
            else:
                with open(fpath, "wb") as download_file:
                    for chunk in response.iter_bytes():
                        download_file.write(chunk)

    def list_runs(self, num: int = 10, only_completed: bool = False):
        """List all the runs on the server.

        - filter by scheduled/completed/in-process?
        """
        url = f"{self.base_url}/runs"
        params: dict = {"num": num}
        if only_completed:
            params["status"] = "completed"
        response = self.client.get(url, params=params)
        return raise_or_return(response)

    def add_daily_schedule(self, checkset: str, start_on: str, assign_topics: int = 0, assign_topics_args: t.Optional[dict] = None):
        """Schedules a periodic evaluation on the server. Specify the checkset to run against it.

        Args:
            checkset: name of the checkset to use
            start_on: date to start the schedule on

        Returns:
            run: information about the schedule along with a unique identifier.
        """
        url = f"{self.base_url}/schedule"
        response = self.client.post(
            url, 
            json={
                "checkset": checkset,
                "start_on": start_on,
                "assign_topics": assign_topics,
                "assign_topics_args": assign_topics_args
            }
        )
        return raise_or_return(response)

    def get_schedule(self, schedule_id: str):
        """Get the status of a schedule.

        Args:
            schedule_id: unique identifier for the schedule.

        Returns:
            run: information about the schedule along with a unique identifier.
        """
        url = f"{self.base_url}/schedule/{schedule_id}"
        response = self.client.get(url)
        return raise_or_return(response)

    def remove_schedule(self, schedule_id: str):
        """Remove a schedule.

        Args:
            schedule_id: unique identifier for the schedule.

        Returns:
            run: information about the schedule along with a unique identifier.
        """
        url = f"{self.base_url}/schedule/{schedule_id}"
        response = self.client.delete(url)
        return raise_or_return(response)

    def list_schedules(self, num: int = 10, active_only: bool = True):
        """List all the schedules on the server."""
        url = f"{self.base_url}/schedules"
        params: dict = {"num": num, "active_only": active_only}
        response = self.client.get(url, params=params)
        return raise_or_return(response)

    def rerun_schedule(
        self, schedule_id: str, start_on: str, end_on: t.Optional[str] = None
    ):
        """Rerun a schedule.
        - New checks added to the checkset are run for all dates
        - Existing checks are run only on the failed rows.

        Args:
            schedule_id: unique identifier for the run.
            start_on: date to start the reruns on
            end_on: date to end the reruns on

        Returns:
            run: information about the schedule along with a unique identifier.
        """
        url = f"{self.base_url}/schedule/{schedule_id}/rerun"
        params = {"start_on": start_on}
        if end_on is not None:
            params["end_on"] = end_on
        response = self.client.put(url, params=params)
        return raise_or_return(response)

    def get_schedule_results_all_dates(self, schedule_id: str, check_name: str) -> list[dict]:
        """Get the results of a schedule.

        Args:
            schedule_id: unique identifier for the schedule.
            check_name: name of the check to get results for.

        Returns:
            run: information about the schedule along with a unique identifier.
        """
        url = f"{self.base_url}/schedule/{schedule_id}/results"
        params: dict = {"check_name": check_name}
        response = self.client.get(url, params=params)
        if not response.is_success:
            logger.error(response.text)
            response.raise_for_status()
        else:
            return pl.read_ndjson(response.content).to_dicts()

    def evaluate(
        self,
        eval_name: str,
        full_dataset: t.Union[list[dict], pl.DataFrame, pd.DataFrame],
        params: t.Union[dict, None] = None,
    ):
        """Run an evaluation on the server.

        NOTE: Internal use only. Use regular uptrain operators to run evaluations,
        """
        url = f"{self.base_url}/evaluate"
        if isinstance(full_dataset, pl.DataFrame):
            full_dataset = full_dataset.to_dicts()
        elif isinstance(full_dataset, pd.DataFrame):
            full_dataset = full_dataset.to_dict(orient="records")

        # send in chunks of 50, so the connection doesn't time out waiting for the server
        results = []

        if params is not None:
            params['uptrain_settings'] = self.settings.dict()
        else:
            params = {}
            params['uptrain_settings'] = self.settings.dict()

        NUM_TRIES = 3
        for i in range(0, len(full_dataset), 20):
            response_json = None
            for try_num in range(NUM_TRIES):
                try:
                    logger.info(
                        f"Sending evaluation request for rows {i} to <{i+20} to the Uptrain server"
                    )
                    response = self.client.post(
                        url,
                        json={
                            "eval_name": eval_name,
                            "dataset": full_dataset[i : i + 20],
                            "params": params,
                        },
                    )
                    response_json = raise_or_return(response)
                    break
                except Exception as e:
                    logger.info("Retrying evaluation request")
                    if try_num == NUM_TRIES - 1:
                        logger.error(f"Evaluation failed with error: {e}")
                        raise e

            if response_json is not None:
                results.extend(response_json)

        return results

    def log_and_evaluate(
        self,
        project_name: str,
        data: t.Union[list[dict], pl.DataFrame, pd.DataFrame],
        checks: list[t.Union[str, Evals, ParametricEval]],
        schema: t.Union[DataSchema, dict[str, str], None] = None,
        metadata: t.Optional[dict[str, t.Any]] = None,
    ):
        """Run an evaluation on the server and log the results.
        NOTE: This api is a bit different than the regular `evaluate` call.

        Args:
            project_name: Name of the project to evaluate on.
            data: Data to evaluate on. Either a Pandas DataFrame or a list of dicts.
            checks: List of checks to evaluate on.
            schema: Schema of the data. Only required if the data attributes aren't typical (question, response, context).
            metadata: Attributes to attach to this dataset. Useful for filtering and grouping in the UI.

        Returns:
            results: List of dictionaries with each data point and corresponding evaluation results.
        """
        url = f"{self.base_url}/log_and_evaluate"

        if isinstance(data, pl.DataFrame):
            data = data.to_dicts()
        elif isinstance(data, pd.DataFrame):
            data = data.to_dict(orient="records")

        if schema is None:
            schema = DataSchema()
        elif isinstance(schema, dict):
            schema = DataSchema(**schema)

        if metadata is None:
            metadata = {}

        checks = [Evals(m) if isinstance(m, str) else m for m in checks]
        for m in checks:
            assert isinstance(m, (Evals, ParametricEval))

        req_attrs, ser_checks = set(), []
        for m in checks:
            if m in [Evals.FACTUAL_ACCURACY, Evals.RESPONSE_COMPLETENESS_WRT_CONTEXT, Evals.RESPONSE_CONSISTENCY]:
                req_attrs.update([schema.question, schema.context, schema.response])
            elif m in [Evals.RESPONSE_RELEVANCE, Evals.RESPONSE_COMPLETENESS, Evals.RESPONSE_CONCISENESS, Evals.PROMPT_INJECTION]:
                req_attrs.update([schema.question, schema.response])
            elif m in [Evals.CONTEXT_RELEVANCE]:
                req_attrs.update([schema.question, schema.context])
            elif m in [Evals.VALID_RESPONSE, Evals.CRITIQUE_LANGUAGE] or isinstance(m, CritiqueTone) or isinstance(m, GuidelineAdherence):
                req_attrs.update([schema.response])
            elif m in [Evals.RESPONSE_ALIGNMENT_WITH_SCENARIO, Evals.RESPONSE_SINCERITY_WITH_SCENARIO]:
                req_attrs.update([schema.question, schema.response, schema.scenario, schema.objective])
            elif isinstance(m, ResponseMatching):
                req_attrs.update([schema.response, schema.ground_truth])
            elif isinstance(m, ConversationSatisfaction):
                req_attrs.update([schema.conversation])

            if isinstance(m, ParametricEval):
                ser_checks.append({"check_name": m.__class__.__name__, **m.dict()})
            elif isinstance(m, Evals):
                ser_checks.append(m.value)
            else:
                raise ValueError(f"Invalid metric: {m}")
        for idx, row in enumerate(data):
            if not req_attrs.issubset(row.keys()):
                raise ValueError(
                    f"Row {idx} is missing required all required attributes for evaluation: {req_attrs}"
                )

        # send in chunks of 50, so the connection doesn't time out waiting for the server
        results = []
        NUM_TRIES, BATCH_SIZE = 3, 50
        for i in range(0, len(data), BATCH_SIZE):
            response_json = None
            for try_num in range(NUM_TRIES):
                try:
                    logger.info(
                        f"Sending evaluation request for rows {i} to <{i+BATCH_SIZE} to the Uptrain server"
                    )
                    response = self.client.post(
                        url,
                        json={
                            "data": data[i : i + BATCH_SIZE],
                            "checks": ser_checks,
                            "metadata": {
                                "project": project_name,
                                "schema": schema.dict(),
                                **metadata,
                                "uptrain_settings": self.settings.dict(),
                            },
                        },
                    )
                    response_json = raise_or_return(response)
                    break
                except Exception as e:
                    logger.info("Retrying evaluation request")
                    if try_num == NUM_TRIES - 1:
                        logger.error(f"Evaluation failed with error: {e}")
                        raise e

            if response_json is not None:
                results.extend(response_json)

        return results


    def evaluate_experiments(
        self,
        project_name: str,
        data: t.Union[list[dict], pl.DataFrame],
        checks: list[t.Union[str, Evals, ParametricEval]],
        exp_columns: list[str],
        schema: t.Union[DataSchema, dict[str, str], None] = None,
        metadata: t.Optional[dict[str, t.Any]] = None,
    ):
        """Evaluate experiments on the server and log the results.

        Args:
            project_name: Name of the experiment to evaluate on.
            data: Data to evaluate on. Either a Pandas DataFrame or a list of dicts.
            checks: List of checks to evaluate on.
            exp_columns: List of columns/keys which denote different experiment configurations.
            schema: Schema of the data. Only required if the data attributes aren't typical (question, response, context).
            metadata: Attributes to attach to this dataset. Useful for filtering and grouping in the UI.

        Returns:
            results: List of dictionaries with each data point and corresponding evaluation results for all the experiments.
        """
        if metadata is None:
            metadata = {}

        metadata.update({'uptrain_experiment_columns': exp_columns})

        if schema is None:
            schema = DataSchema()
        elif isinstance(schema, dict):
            schema = DataSchema(**schema)

        results = self.log_and_evaluate(
            project_name=project_name,
            data=data,
            checks=checks,
            schema=schema,
            metadata=metadata,
        )

        results = pl.DataFrame(results)
        all_cols = set(results.columns)
        value_cols = list(all_cols - set([schema.question] + exp_columns))
        exp_results = results.pivot(values=value_cols, index=schema.question, columns=exp_columns)
        exp_results = exp_results.to_dicts()
        return exp_results


    def download_project_eval_results(self, project_name: str, fpath: str):
        """Fetch all the evaluation results for a project.

        Args:
            project_name: Name of the project to fetch results for.
            fpath: Path to save the results to.
        """
        url = f"{self.base_url}/evaluation_results/{project_name}"
        with self.client.stream("GET", url) as response:
            if not response.is_success:
                logger.error(response.text)
                response.raise_for_status()
            else:
                with open(fpath, "w") as download_file:
                    for chunk in response.iter_text():
                        download_file.write(chunk)



"""
Implement checks to test language quality.

This module provides the `GrammarScore` class, which allows evaluating the grammatical correctness of sentences using
the OpenAI GPT-3.5-turbo language model. It generates a grammar score for each input sentence on a scale of 0 to 100.

"""

from __future__ import annotations
import re
import typing as t

from loguru import logger
import polars as pl

if t.TYPE_CHECKING:
    from uptrain.framework import Settings
from uptrain.operators.base import *
from uptrain.operators.language.llm import LLMMulticlient, Payload

__all__ = ["GrammarScore"]


@register_op
class GrammarScore(ColumnOp):
    """
    Operator to test the grammatical correctness of sentences using the OpenAI GPT-3.5-turbo language model.

    Attributes:
        col_in_text (str): The name of the input column containing the sentences to evaluate.
        col_out (str): The name of the output column containing the grammar scores.

    Returns:
        dict: A dictionary containing the grammar scores for each input sentence.

    """

    col_in_text: str = "text"
    col_out: str = "grammar_score"

    def setup(self, settings: t.Optional[Settings] = None):
        self._api_client = LLMMulticlient(settings=settings)
        self._settings = settings
        return self

    def _make_payload(self, id: t.Any, text: str) -> Payload:
        data={
                "model": "gpt-3.5-turbo",
                "messages": [
                    {
                        "role": "system",
                        "content": "You are a grammatical correctness evaluator who produces only a number and no explanation.",
                    },
                    {
                        "role": "user",
                        "content": "Score following sentence on grammatical correctness on a scale of 0 to 100: \n\n {statement}".format(
                            statement=text
                        ),
                    },
                ],
        }
        if self._settings.seed is not None:
            data["seed"] = self._settings.seed
            
        return Payload(
            data = data,
            metadata={"index": id},
        )

    def run(self, data: pl.DataFrame) -> TYPE_TABLE_OUTPUT:
        text_ser = data.get_column(self.col_in_text)
        input_payloads = [
            self._make_payload(idx, text) for idx, text in enumerate(text_ser)
        ]
        output_payloads = self._api_client.fetch_responses(input_payloads)

        results = []
        for res in output_payloads:
            assert (
                res is not None
            ), "Response should not be None, we should've handled exceptions beforehand."
            idx = res.metadata["index"]
            if res.error is not None:
                logger.error(
                    f"Error when processing payload at index {idx}: {res.error}"
                )
                results.append((idx, None))
            else:
                resp_text = res.response.choices[0].message.content
                number = int(re.findall(r"\d+", resp_text)[0])
                results.append((idx, number))

        result_scores = pl.Series(
            [val for _, val in sorted(results, key=lambda x: x[0])]
        )
        return {"output": data.with_columns([result_scores.alias(self.col_out)])}



from typing import Any, Dict

from uptrain.v0.core.classes.visuals import AbstractVisual
from uptrain.v0.constants import PlotType, Visual


class Plot(AbstractVisual):
    """Class to expose the plotting API."""

    visual_type = Visual.PLOT

    def base_init(self, framework, check: Dict[str, Any]) -> None:
        """Base init for the Plot class.
        
        Parameters
        ----------
        framework
            UpTrain Framework object
        check
            Configuration dictionary for the plot
            The dictionary must contain the following keys:
                plot: The type of plot to use. Must be one of the following:
                    uptrain.PlotType.BAR_CHART
                    uptrain.PlotType.LINE_CHART
                    uptrain.PlotType.HISTOGRAM
                plot_name: The name of the plot. Defaults to the plot type.
                dashboard_name: The name of the dashboard. Defaults to "Plot".
            
            If the plot type is uptrain.PlotType.BAR_CHART:
                x_feature_name: The name of the feature to use for the x-axis.
                y_feature_name: The name of the feature to use for the y-axis.
            
            If the plot type is uptrain.PlotType.LINE_CHART:
                x_feature_name: The name of the feature to use for the x-axis.
                y_feature_name: The name of the feature to use for the y-axis.
            
            If the plot type is uptrain.PlotType.HISTOGRAM:
                feature_name: The name of the feature to use for the histogram.
        """

        self.framework = framework
        self.plot = check["plot"]
        self.plot_name = check.get("plot_name", self.plot)
        self.dashboard_name = check.get("dashboard_name", "Plot")

        if self.plot == PlotType.BAR_CHART:
            self.x_feature_name = check.get("x_feature_name", None)
            self.y_feature_name = check.get("y_feature_name", None)
            self._chart = self._bar_chart

            if self.x_feature_name is None and self.y_feature_name is None:
                raise Exception(
                    "Both x_feature_name and y_feature_name are None - atleast one must be not None"
                )
        elif self.plot == PlotType.LINE_CHART:
            self.x_feature_name = check.get("x_feature_name", None)
            self.y_feature_name = check.get("y_feature_name", None)
            self.count = 0
            self._chart = self._line_chart

            if self.x_feature_name is None and self.y_feature_name is None:
                raise Exception(
                    "Both x_feature_name and y_feature_name are None - atleast one must be not None"
                )
        elif self.plot == PlotType.HISTOGRAM:
            self.feature_name = check.get("feature_name", None)
            self._chart = self._histogram_chart

            if self.feature_name is None:
                raise Exception("feature_name is None")
        else:
            raise Exception("Plot Type is not supported")

    def base_check(self, inputs, outputs, gts=None, extra_args={}) -> None:
        self._chart(inputs, outputs, gts, extra_args)

    def _bar_chart(self, inputs, outputs, gts=None, extra_args={}) -> None:
        """Function to plot a bar chart."""

        # Check if the provided inputs were for the bar chart
        # TODO: refactor and make the checks better
        if not (
            self.x_feature_name in inputs.keys() or self.y_feature_name in inputs.keys()
        ):
            return
        
        length = (
            len(inputs[self.x_feature_name])
            if self.x_feature_name is not None
            else len(inputs[self.y_feature_name])
        )
        
        if self.x_feature_name is None:
            x_features = list(range(self.count, self.count + length))
            self.count += length
        else:
            x_features = inputs[self.x_feature_name]
        if self.y_feature_name is None:
            y_features = list(range(self.count, self.count + length))
            self.count += length
        else:
            y_features = inputs[self.y_feature_name]
        
        bars = inputs["bars"]
        data = {bar: {} for bar in set(bars)}
        
        for i in range(length):
            data[bars[i]].update({x_features[i]: y_features[i]})
        self.framework.log_handler.add_bar_graphs(
            plot_name=self.plot_name, data=data, dashboard_name=self.dashboard_name
        )

    def _histogram_chart(self, inputs, outputs, gts=None, extra_args={}) -> None:
        """Function to plot a histogram chart."""

        # Check if the provided inputs were for the histogram chart
        # TODO: refactor and make the checks better
        if not (self.feature_name in inputs.keys()):
            return
        self.log_handler.add_histogram(
            plot_name=self.plot_name,
            data=inputs[self.feature_name],
            dashboard_name=self.dashboard_name,
            file_name=f"{self.plot_name}_{self.feature_name}",
        )

    def _line_chart(self, inputs, outputs, gts=None, extra_args={}) -> None:
        """Function to plot a line chart."""
        
        # Check if the provided inputs were for the line chart
        # TODO: refactor and make the checks better
        if not (
            self.x_feature_name in inputs.keys() or self.y_feature_name in inputs.keys()
        ):
            return
        
        length = (
            len(inputs[self.y_feature_name])
            if self.x_feature_name is None
            else len(inputs[self.x_feature_name])
        )
        
        if self.x_feature_name is None:
            x_features = list(range(self.count, self.count + length))
            self.count += length
        else:
            x_features = inputs[self.x_feature_name]
        if self.y_feature_name is None:
            y_features = list(range(self.count, self.count + length))
            self.count += length
        else:
            y_features = inputs[self.y_feature_name]
        
        y_feature_name = (
            f"y_{self.y_feature_name}" if self.y_feature_name is not None else "y_value"
        )
        
        for i in range(length):
            self.framework.log_handler.add_scalars(
                plot_name=self.plot_name,
                dictn={y_feature_name: y_features[i]},
                count=x_features[i],
                dashboard_name=self.dashboard_name,
                file_name=f"{self.plot_name}_{self.x_feature_name}_{self.y_feature_name}",
            )



from __future__ import annotations
from typing import TYPE_CHECKING, Union
import numpy as np

from uptrain.v0.core.lib.helper_funcs import make_2d_np_array
from uptrain.v0.core.classes.distances import DistanceResolver
from uptrain.v0.core.classes.statistics import AbstractStatistic
from uptrain.v0.core.classes.measurables import MeasurableResolver
from uptrain.v0.constants import Statistic
from uptrain.v0.core.lib.cache import make_cache_container

if TYPE_CHECKING:
    from uptrain.v0.core.classes.logging.new_log_handler import (
        LogHandler as NewLogHandler,
        LogWriter,
    )
    from uptrain.v0.core.classes.logging.log_handler import LogHandler

# from uptrain.v0.core.classes.algorithms import Clustering
# from uptrain.v0.core.lib.algorithms import estimate_earth_moving_cost


class Convergence(AbstractStatistic):
    log_handler: Union["LogHandler", "NewLogHandler"]
    log_writers: list["LogWriter"]
    dashboard_name = "convergence_stats"
    statistic_type = Statistic.CONVERGENCE_STATS

    def base_init(self, fw, check):
        self.aggregate_measurable = MeasurableResolver(check["aggregate_args"]).resolve(
            fw
        )
        self.count_measurable = MeasurableResolver(check["count_args"]).resolve(fw)
        self.reference = check["reference"]
        self.distance_types = check["distance_types"]
        self.count_checkpoints = list(sorted(set([0] + check["count_checkpoints"])))

        # setup a cache to store interim state for aggregates
        props_to_store = {
            "ref_embedding": np.ndarray,
            "prev_count": int,
            "first_checkpoint": int,
        }
        self.cache = make_cache_container(fw, props_to_store)

        self.dist_classes = [DistanceResolver().resolve(x) for x in self.distance_types]
        if hasattr(self.log_handler, "make_logger"):
            self.log_writers = [
                self.log_handler.make_logger(
                    self.dashboard_name, distance_type + "_" + str(self.reference)
                )
                for distance_type in self.distance_types
            ]  # get handles to log writers for each distance type
        else:
            self.log_writers = []

    def base_check(self, inputs, outputs, gts=None, extra_args={}):
        all_models = [
            x.compute_and_log(inputs, outputs, gts=gts, extra=extra_args)
            for x in self.model_measurables
        ]
        all_features = [
            x.compute_and_log(inputs, outputs, gts=gts, extra=extra_args)
            for x in self.feature_measurables
        ]
        vals = self.measurable.compute_and_log(
            inputs, outputs, gts=gts, extra=extra_args
        )
        aggregate_ids = self.aggregate_measurable.compute_and_log(
            inputs, outputs, gts=gts, extra=extra_args
        )
        counts = self.count_measurable.compute_and_log(
            inputs, outputs, gts=gts, extra=extra_args
        )

        # read previous state from the cache
        # - ref_embs_cache: the reference embedding for each aggregate id
        # - prev_count_cache: the previous view count seen for each aggregate id
        # - first_crossed_checkpoint_cache: the first checkpoint that was crossed for each aggregate id
        [
            ref_embs_cache,
            prev_count_cache,
            first_crossed_checkpoint_cache,
        ] = self.cache.fetch_col_values_for_ids(
            np.asarray(aggregate_ids),
            ["ref_embedding", "prev_count", "first_checkpoint"],
        )

        set_ids_to_cache = set()  # track which ids we need to update in the cache
        for idx in range(len(aggregate_ids)):
            is_model_invalid = sum(
                [
                    all_models[jdx][idx] not in self.allowed_model_values[jdx]
                    for jdx in range(len(self.allowed_model_values))
                ]
            )
            if is_model_invalid:
                continue

            active_id = aggregate_ids[idx]
            prev_count = prev_count_cache.get(active_id, None)
            curr_count = counts[idx]
            ref_emb = ref_embs_cache.get(active_id, None)
            curr_emb = vals[idx]

            # Four possible cases:
            # 1. prev_count = None: first time we see this aggregate id, so we skip processing it. Save the current state in the cache.
            # 2. prev_count > max(checkpoints): we skip processing this row altogether since all checkpoints have been crossed.
            # 3. prev_count > curr_count: this row is out of order, so we skip processing it.
            # 4. curr_count > prev_count: we have seen this aggregate id before, so we process it. Save the current state in the cache.
            if prev_count is None:
                crossed_checkpoint = max(
                    cp for cp in self.count_checkpoints if cp <= curr_count
                )

                set_ids_to_cache.add(active_id)
                prev_count_cache[active_id] = curr_count
                ref_embs_cache[active_id] = curr_emb
                first_crossed_checkpoint_cache[active_id] = crossed_checkpoint
            elif prev_count > self.count_checkpoints[-1]:
                continue
            elif prev_count > curr_count:
                continue
            else:
                last_crossed_checkpoint = None
                for cp in reversed(self.count_checkpoints):
                    if prev_count > cp:
                        break  # we are already past this checkpoint or those smaller than it
                    elif (prev_count <= cp) and (cp < curr_count):
                        last_crossed_checkpoint = cp
                        break

                set_ids_to_cache.add(active_id)
                prev_count_cache[active_id] = curr_count

                if last_crossed_checkpoint is None:
                    # no new checkpoint crossed, so we just update the last count seen and move on
                    continue
                else:
                    # new checkpoint crossed and we have a reference embedding, so compute the statistics.
                    # Update the reference embedding if necessary.
                    assert ref_emb is not None
                    if self.reference == "running":
                        ref_embs_cache[active_id] = curr_emb

                    # compute the statistics
                    distances = dict(
                        zip(
                            self.distance_types,
                            [
                                x.compute_distance(
                                    make_2d_np_array(curr_emb),
                                    make_2d_np_array(ref_emb),
                                )[0]
                                for x in self.dist_classes
                            ],
                        )
                    )

                    models = dict(
                        zip(
                            self.model_names,
                            [
                                all_models[jdx][idx]
                                for jdx in range(len(self.model_names))
                            ],
                        )
                    )
                    features = dict(
                        zip(
                            self.feature_names,
                            [
                                all_features[jdx][idx]
                                for jdx in range(len(self.feature_names))
                            ],
                        )
                    )

                    if len(self.log_writers) > 0:
                        for k, distance_type in enumerate(self.distance_types):
                            self.log_writers[k].log(
                                {
                                    "check": distances[distance_type],
                                    self.aggregate_measurable.feature_name: aggregate_ids[idx],  # type: ignore
                                    self.count_measurable.feature_name: last_crossed_checkpoint,  # type: ignore
                                    **models,
                                    **features,
                                }
                            )
                    else:
                        for distance_key in list(distances.keys()):
                            plot_name = distance_key + " " + str(self.reference)
                            this_data = list(
                                np.reshape(np.array(distances[distance_key]), -1)
                            )
                            self.log_handler.add_histogram(
                                plot_name,
                                this_data,
                                self.dashboard_name,
                                models=[models] * len(this_data),
                                features=[features] * len(this_data),
                                file_name=str(last_crossed_checkpoint),
                            )

        if len(set_ids_to_cache) > 0:
            # save the current state in the cache
            list_ids = list(set_ids_to_cache)
            list_ref_embs, list_counts, list_first_checkpoints = [], [], []
            for _id in list_ids:
                list_ref_embs.append(ref_embs_cache[_id])
                list_counts.append(prev_count_cache[_id])
                list_first_checkpoints.append(first_crossed_checkpoint_cache[_id])

            self.cache.upsert_ids_n_col_values(
                np.asarray(list_ids),
                {
                    "ref_embedding": np.asarray(list_ref_embs),
                    "prev_count": np.asarray(list_counts),
                    "first_checkpoint": np.asarray(list_first_checkpoints),
                },
            )



import os
from scipy.spatial.distance import cosine
import subprocess
import json

import uptrain.v0 as v0
from gensim.models import Word2Vec


def test_recommender_bias():
    data_file = "sigir_data.json"
    remote_url = "https://oodles-dev-training-data.s3.amazonaws.com/sigir_data.json"
    if not os.path.exists(data_file):
        try:
            # Most Linux distributions have Wget installed by default.
            # Below command is to install wget for MacOS
            wget_installed_ok = subprocess.call("brew install wget", shell=True, stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT)
            print("Successfully installed wget")
        except:
            dummy = 1
        try:
            if not os.path.exists("sigir_data.json"):
                file_downloaded_ok = subprocess.call("wget " + remote_url, shell=True, stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT)
                print("Data downloaded")
        except:
            print("Could not load training data")
            print("Please follow following steps to manually download data")
            print("Step 1: Open in browser: https://oodles-dev-training-data.s3.amazonaws.com/sigir_data.json")
            print("Step 2: Download and move the file to example location (i.e. uptrain/examples/3_shopping_cart_recommendation/")
    else:
        print("Data file exists. Skipping download.")


    with open('sigir_data.json') as f:
        data = json.load(f)

    x_train_sku = [[e['product_sku'] for e in s] for s in data['x_train']]
    model = Word2Vec(sentences=x_train_sku, vector_size=48, epochs=15).wv

    def cosine_dist_init(self):
        self.cos_distances = []
        self.model = model

    def cosine_distance_check(self, inputs, outputs, gts=None, extra_args={}):
        for output, gt in zip(outputs, gts):
            if (not output) or (not gt):
                continue
            y_preds = output[0]
            y_gt = gt
            try:
                vector_test = self.model.get_vector(y_gt)
            except:
                vector_test = []
            vector_pred = self.model.get_vector(y_preds)
            if len(vector_pred)>0 and len(vector_test)>0:
                cos_dist = cosine(vector_pred, vector_test)
                self.cos_distances.append(cos_dist)
                self.log_handler.add_histogram('cosine_distance', self.cos_distances, self.dashboard_name)

    x_test = data['x_test']
    y_test = data['y_test']
    inference_batch_size = 10

    def model_predict(model, x_test_batch):
        """
        Implement the model prediction function. 
        
        :model: Word2Vec model learned from user shopping sessions
        :x_test_batch: list of lists, each list being the content of a cart
        
        :return: the predictions returned by the model are the top-K
        items suggested to complete the cart.

        """
        predictions = []
        for _x in x_test_batch:
            key_item = _x[0]['product_sku']
            nn_products = model.most_similar(key_item, topn=10) if key_item in model else None
            if nn_products:
                predictions.append([_[0] for _ in nn_products])
            else:
                predictions.append([])

        return predictions


    cfg = {
        "checks": [
            # Check to monitor the hit-rate of the model 
            # (i.e., if any recommended item was selected)
            {
                'type': v0.Monitor.CONCEPT_DRIFT,
                'algorithm': v0.DataDriftAlgo.DDM,
                'warn_thres': 2,
                'alarm_thres': 3,
                "measurable_args": {
                    'type': v0.MeasurableType.REC_HIT_RATE,
                    },
            },
            # Define a check on popularity bias
            {
                'type': v0.Monitor.POPULARITY_BIAS,
                'algorithm': v0.BiasAlgo.POPULARITY_BIAS,
                'sessions': x_train_sku,   
            },
            # Custom monitor
            {
                'type': v0.Monitor.CUSTOM_MONITOR,
                'initialize_func': cosine_dist_init,
                'check_func': cosine_distance_check,
                'need_gt': True,
                'dashboard_name': 'cosine_distance'
            },
        ], 
        "retraining_folder": 'uptrain_smart_data', 
    }

    framework = v0.Framework(cfg)


    for i in range(int(len(x_test)/inference_batch_size)):
        # Define input in the format understood by the UpTrain framework
        inputs = {"feats": x_test[i*inference_batch_size:(i+1)*inference_batch_size]}
        
        # Do model prediction
        preds = model_predict(model, inputs['feats'])

        # Log input and output to framework
        ids = framework.log(inputs=inputs, outputs=preds)
        
        # Getting ground truth and logging to framework
        y_test_batch = y_test[i*inference_batch_size:(i+1)*inference_batch_size]
        gts = [y[0]['product_sku'] for y in y_test_batch]
        framework.log(identifiers=ids, gts=gts)
        


import os
import subprocess
import zipfile
import numpy as np
from . import read_json, write_json 

def get_data_from_remote():

    data_dir = "data"
    remote_url = "https://oodles-dev-training-data.s3.amazonaws.com/data.zip"
    orig_training_file = 'data/training_data.json'
    if not os.path.exists(data_dir):
        try:
            # Most Linux distributions have Wget installed by default.
            # Below command is to install wget for MacOS
            wget_installed_ok = subprocess.call("brew install wget", shell=True, stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT)
            print("Successfully installed wget")
        except:
            dummy = 1
        try:
            if not os.path.exists("data.zip"):
                file_downloaded_ok = subprocess.call("wget " + remote_url, shell=True, stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT)
                print("Data downloaded")
            with zipfile.ZipFile("data.zip", 'r') as zip_ref:
                zip_ref.extractall("./")
            full_training_data = read_json(orig_training_file)
            np.random.seed(1)
            np.random.shuffle(full_training_data)
            reduced_training_data = full_training_data[0:1000]
            write_json(orig_training_file, reduced_training_data)
            print("Prepared Example Dataset")
            os.remove("data.zip")
        except Exception as e:
            print(e)
            print("Could not load training data")



"""Utility routines to help us use Duckdb, as storage for intermediate state when 
computing aggregations. 
"""

from __future__ import annotations
import os
from typing import Any, Type, Union
import uuid

import duckdb
import numpy as np
import pyarrow as pa

# -----------------------------------------------------------
# utility routines for converting between Arrow and Numpy
# -----------------------------------------------------------


def array_np_to_arrow(arr: np.ndarray) -> pa.Array:
    assert arr.ndim in (1, 2), "Only 1D and 2D arrays are supported."
    if arr.ndim == 1:
        return pa.array(arr)
    else:
        dim1, dim2 = arr.shape
        return pa.ListArray.from_arrays(
            np.arange(0, (dim1 + 1) * dim2, dim2), arr.ravel()
        )


def array_arrow_to_np(arr: pa.Array) -> np.ndarray:
    if isinstance(arr, pa.ChunkedArray):
        arr = arr.combine_chunks()

    if not pa.types.is_list(arr.type):
        return arr.to_numpy()  # assume a 1D array
    else:
        dim1 = len(arr)  # assume a 2D array
        return np.asarray(arr.values.to_numpy()).reshape(dim1, -1)


def arrow_batch_to_table(batch_or_tbl: Union[pa.Table, pa.RecordBatch]) -> pa.Table:
    if not isinstance(batch_or_tbl, pa.Table):
        return pa.Table.from_batches([batch_or_tbl])
    else:
        return batch_or_tbl


def table_arrow_to_np_arrays(tbl: pa.Table, cols: list[str]) -> list[np.ndarray]:
    return [array_arrow_to_np(tbl[c]) for c in cols]


def np_arrays_to_arrow_table(arrays: list[np.ndarray], cols: list[str]) -> pa.Table:
    return pa.Table.from_pydict(
        {c: array_np_to_arrow(arr) for c, arr in zip(cols, arrays)}
    )


# -----------------------------------------------------------
# State cache
# -----------------------------------------------------------


class DuckDBStateCache:
    """A state cache that uses DuckDB to store intermediate state."""

    def __init__(self, fw: Any, columns: dict[str, Type]):
        db_dir = os.path.join(fw.fold_name, "dbs")
        os.makedirs(db_dir, exist_ok=True)
        self.conn = duckdb.connect(os.path.join(db_dir, str(uuid.uuid4()) + ".db"))

        # create the table
        cols = {"id": "LONG PRIMARY KEY"}
        for col, col_type in columns.items():
            if col_type == np.ndarray:
                cols[col] = "FLOAT[]"
            elif col_type == float:
                cols[col] = "FLOAT"
            elif col_type == int:
                cols[col] = "LONG"
            elif col_type == str:
                cols[col] = "VARCHAR"
            else:
                raise ValueError(f"Unsupported type {col_type}")
        str_cols = ",".join([f"{col} {col_type}" for col, col_type in cols.items()])
        self.conn.execute(f"CREATE TABLE store ({str_cols})")

    def upsert_ids_n_col_values(
        self, ids: np.ndarray, col_values: dict[str, np.ndarray]
    ):
        tbl = pa.Table.from_pydict(
            {
                "id": array_np_to_arrow(ids),
                **{
                    col: array_np_to_arrow(values) for col, values in col_values.items()
                },
            }
        )
        # upserts don't work for List data types, so we do it in steps
        # conn.execute("INSERT OR REPLACE INTO store SELECT id, value FROM tbl")
        self.conn.execute(f"DELETE FROM store WHERE id IN (SELECT id FROM tbl)")
        str_col_names = ",".join(col_values.keys())
        self.conn.execute(f"INSERT INTO store SELECT id, {str_col_names} FROM tbl")

    def fetch_col_values_for_ids(
        self, ids: np.ndarray, col_names: list[str]
    ) -> list[dict]:
        id_tbl = pa.Table.from_pydict({"id": array_np_to_arrow(ids)})
        str_col_names = ",".join(col_names)
        self.conn.execute(
            f"SELECT id, {str_col_names} FROM store WHERE id IN (SELECT id FROM id_tbl)"
        )

        value_tbl = self.conn.fetch_arrow_table()
        if len(value_tbl) == 0:
            return [{} for _ in col_names]
        else:
            output = []
            _ids = array_arrow_to_np(value_tbl["id"])
            for col in col_names:
                _values = array_arrow_to_np(value_tbl[col])
                output.append({_id: value for _id, value in zip(_ids, _values)})
            return output



"""IO operators to read/write from/to Duckdb."""

from __future__ import annotations
import typing as t

import polars as pl

if t.TYPE_CHECKING:
    from uptrain.framework import Settings
from uptrain.operators.base import *
from uptrain.utilities import lazy_load_dep


duckdb = lazy_load_dep("duckdb", "duckdb")


@register_op
class DuckDBReader(TransformOp):
    """Read data from a Duckdb table.

    Attributes:
        fpath (str): Path to the Duckdb file.
        query (str): Query to run against the duckdb database.
        col_timestamp (str): Column name to use as the timestamp column. Only used in the context of monitoring.

    Example:
        ```python
        from uptrain.operators.io import DuckDBReader

        reader = DuckDBReader(
            fpath="data/duckdb.db",
            query="SELECT * FROM my_table",
            col_timestamp="timestamp",
        )
        output = reader.setup().run()["output"]
        ```
    """

    fpath: str
    query: str
    col_timestamp: str = "timestamp"

    def setup(self, settings: Settings):
        self._conn = duckdb.connect(self.fpath)
        return self

    def run(self) -> TYPE_TABLE_OUTPUT:
        res = self._conn.query(self.query).to_arrow_table()
        return {"output": pl.from_arrow(res)}



import os
import tensorflow as tf
import numpy as np

from helper_files import KpsDataset, read_json

import random

random.seed(0)
tf.random.set_seed(0)


def make_model(input_shape=34):
    kernel_initializer = tf.keras.initializers.GlorotUniform(seed=10)
    bias_initializer = tf.keras.initializers.zeros()

    inputs = tf.keras.Input(shape=input_shape)
    x = tf.keras.layers.Dense(
        18,
        activation="relu",
        kernel_initializer=kernel_initializer,
        bias_initializer=bias_initializer,
    )(inputs)
    outputs = tf.keras.layers.Dense(
        1,
        activation="sigmoid",
        kernel_initializer=kernel_initializer,
        bias_initializer=bias_initializer,
    )(x)
    return tf.keras.Model(inputs, outputs)


def get_accuracy_tf(testing_file, model_save_name, model_dir="trained_models_tf/"):
    testing_dataset = KpsDataset(
        testing_file,
        batch_size=len(read_json(testing_file)),
        shuffle=False,
        augmentations=False,
        is_test=True,
    )
    model = tf.keras.models.load_model(model_dir + model_save_name)
    print("Evaluating on ", len(read_json(testing_file)), " data-points")

    pred_classes = []
    gt_classes = []
    for elem in testing_dataset:
        preds = model.predict(elem[0]["kps"])
        pred_class = [int(round(float(x))) for x in preds]
        pred_classes.extend(pred_class)
        gt_classes.extend(list(elem[1]))

    accuracy = np.sum(np.array(gt_classes) == np.array(pred_classes)) / len(
        pred_classes
    )
    return accuracy


def train_model_tf(training_file, model_save_name, model_dir="trained_models_tf/"):
    print(
        "Training on: ",
        training_file,
        " which has ",
        len(read_json(training_file)),
        " data-points",
    )
    model_loc = model_dir + model_save_name
    if os.path.exists(model_loc):
        print("Trained model exists. Skipping training again.")
        return

    model = make_model(input_shape=34)
    model.compile(
        optimizer=tf.keras.optimizers.Adam(learning_rate=5e-4),
        loss="binary_crossentropy",
        metrics=tf.keras.metrics.BinaryAccuracy(),
    )

    training_dataset = KpsDataset(training_file, shuffle=True, augmentations=True)
    x = [x[0] for x in training_dataset]
    y = [y[1] for y in training_dataset]
    dataset = tf.data.Dataset.from_tensor_slices((x, y))
    model.fit(dataset, epochs=10)
    model.save(model_loc)
    print("Model saved at: ", model_loc)



"""IO operators to read/write from/to Bigquery."""

from __future__ import annotations
import typing as t
import io

import polars as pl

if t.TYPE_CHECKING:
    from uptrain.framework import Settings
from uptrain.operators.base import *
from uptrain.utilities import lazy_load_dep


bigquery = lazy_load_dep("google.cloud.bigquery", "google-cloud-bigquery")


@register_op
class BigQueryReader(TransformOp):
    """Read data from a bigquery table.

    NOTE: To use this operator, you must include the GCP service account credentials in
    the settings, using the key `gcp_service_account_credentials`.

    Attributes:
        query (str): Query to run against the BigQuery table.
        col_timestamp (str): Column name to use as the timestamp column. Only used in the context of monitoring.

    Example:
        ```python
        from uptrain.operators.io import BigQueryReader

        query = "SELECT * FROM `bigquery-public-data.samples.shakespeare` LIMIT 10"
        reader = BigQueryReader(
            query=query,
            col_timestamp="timestamp"
        )
        output = reader.setup().run()["output"]
        ```
    """

    query: str
    col_timestamp: str = "timestamp"
    limit_rows: int | None = None

    def setup(self, settings: Settings):
        from google.oauth2 import service_account

        gcp_sa_creds = settings.check_and_get("gcp_service_account_credentials")
        credentials = service_account.Credentials.from_service_account_info(
            gcp_sa_creds
        )
        self._client = bigquery.Client(credentials=credentials)
        return self

    def run(self) -> TYPE_TABLE_OUTPUT:
        if self.limit_rows is None:
            query_job = self._client.query(self.query)
        else:
            query_job = self._client.query(self.query + " LIMIT " + str(self.limit_rows))
        rows = query_job.result()
        return {"output": pl.from_arrow(rows.to_arrow())}


@register_op
class BigQueryWriter(TransformOp):
    """Write data to a bigquery table.

    NOTE: To use this operator, you must include the GCP service account credentials in
    the settings, using the key `gcp_service_account_credentials`.

    Attributes:
        project_name (str): Name of the project to add to.
        table_name (str): Name of the table to add to. Final destination is project_name + table_name.
        update_query (str): Optional query if one want to update an existing table with the evaluation results.
        delete_query (str): Optional query if one wants to delete the newly created table (after update).

    Example:
        ```python
        from uptrain.operators.io import BigQueryWriter

        writer = BigQueryWriter(
            project_name="project",
            table_name="dataset.temp_table",
            update_query="UPDATE `{project}.{dataset}.{table}` AS target SET target.score = temp.score FROM `{project}.{dataset}.temp_table` AS temp WHERE target.id = temp.id;
            delete_query="DROP TABLE `project.dataset.temp_table`"
        )
        writer.setup(settings).run()
        ```
    """

    project_name: str
    table_name: str
    update_query: t.Union[str, None] = None
    delete_query: t.Union[str, None] = None

    def setup(self, settings: Settings):
        from google.oauth2 import service_account

        gcp_sa_creds = settings.check_and_get("gcp_service_account_credentials")
        credentials = service_account.Credentials.from_service_account_info(
            gcp_sa_creds
        )
        self._client = bigquery.Client(credentials=credentials)
        return self

    def run(self, data: pl.DataFrame) -> TYPE_TABLE_OUTPUT:

        with io.BytesIO() as stream:
            data.write_parquet(stream)
            stream.seek(0)
            query_job = self._client.load_table_from_file(
                stream,
                destination=self.table_name,
                project=self.project_name,
                job_config=bigquery.LoadJobConfig(
                    source_format=bigquery.SourceFormat.PARQUET,
                ),
            )
        query_job.result()  # Waits for the job to complete

        if self.update_query is not None:
            query_job = self._client.query(self.update_query)
            query_job.result()

        if self.delete_query is not None:
            query_job = self._client.query(self.delete_query)
            query_job.result()

        return {"output": data}


from __future__ import annotations
from typing import TYPE_CHECKING, Union
import numpy as np

from uptrain.v0.core.classes.statistics import AbstractStatistic
from uptrain.v0.core.classes.distances import DistanceResolver
from uptrain.v0.core.classes.measurables import MeasurableResolver
from uptrain.v0.constants import Statistic
from uptrain.v0.core.lib.cache import make_cache_container


if TYPE_CHECKING:
    from uptrain.v0.core.classes.logging.new_log_handler import (
        LogHandler as NewLogHandler,
        LogWriter,
    )
    from uptrain.v0.core.classes.logging.log_handler import LogHandler


class Distance(AbstractStatistic):
    log_handler: Union["LogHandler", "NewLogHandler"]
    log_writers: list["LogWriter"]
    dashboard_name = "distance"
    statistic_type = Statistic.DISTANCE

    def base_init(self, fw, check):
        self.aggregate_measurable = MeasurableResolver(check["aggregate_args"]).resolve(
            fw
        )
        self.count_measurable = MeasurableResolver(check["count_args"]).resolve(fw)
        self.reference = check["reference"]
        self.distance_types: list = check["distance_types"]
        self.dist_classes = [DistanceResolver().resolve(x) for x in self.distance_types]

        # setup a cache to store interim state for aggregates
        attrs_to_store = {"ref_embedding": np.ndarray}
        self.cache = make_cache_container(fw, attrs_to_store)

        # get handles to log writers for each distance type
        if hasattr(self.log_handler, "make_logger"):
            self.log_writers = [
                self.log_handler.make_logger(
                    self.dashboard_name, distance_type + "_" + str(self.reference)
                )
                for distance_type in self.distance_types
            ]
        else:
            self.log_writers = []

    def base_check(self, inputs, outputs=None, gts=None, extra_args={}):
        all_models = [
            x.compute_and_log(inputs, outputs, gts=gts, extra=extra_args)
            for x in self.model_measurables
        ]
        all_features = [
            x.compute_and_log(inputs, outputs, gts=gts, extra=extra_args)
            for x in self.feature_measurables
        ]

        vals_all = self.measurable.compute_and_log(
            inputs, outputs, gts=gts, extra=extra_args
        )
        aggregate_ids_all = self.aggregate_measurable.compute_and_log(
            inputs, outputs, gts=gts, extra=extra_args
        )
        counts_all = self.count_measurable.compute_and_log(
            inputs, outputs, gts=gts, extra=extra_args
        )

        # each batch has rows corresponding to all model values, so we filter out the ones worked on elsewhere
        select_idxs = []
        for idx in range(len(aggregate_ids_all)):
            is_model_invalid = sum(
                [
                    all_models[jdx][idx] not in self.allowed_model_values[jdx]
                    for jdx in range(len(self.allowed_model_values))
                ]
            )
            if not is_model_invalid:
                select_idxs.append(idx)
        if len(select_idxs) == 0:
            return  # nothing for us to do here

        # get the subset of id, counts and values we need to work with
        aggregate_ids = aggregate_ids_all[select_idxs]
        vals = vals_all[select_idxs, :]
        counts = counts_all[select_idxs]
        [ref_embs_cache] = self.cache.fetch_col_values_for_ids(
            np.asarray(aggregate_ids), ["ref_embedding"]
        )  # NOTE: this gives us a unique value for each id, as observed from previous batches

        # Get the reference embedding for each row in this batch.
        # - for initial diff, you want the first value seen for this id - either cached or from this batch
        # - for running diff, you want the previous value seen for this id - either cached or from this batch
        ref_embs_prev, list_ref_vals = {}, []
        for i, key in enumerate(aggregate_ids):
            value_from_cache = ref_embs_cache.get(key, vals[i])
            value_prev_seen_this_batch = ref_embs_prev.get(key, vals[i])
            if self.reference == "running":
                list_ref_vals.append(value_prev_seen_this_batch)
                ref_embs_prev[key] = vals[i]
            else:
                list_ref_vals.append(value_from_cache)
                ref_embs_prev[key] = value_from_cache

        ref_vals = np.vstack(list_ref_vals)
        if len(ref_vals.shape) == 1:
            ref_vals = np.expand_dims(ref_vals, 0)
        distances = dict(
            zip(
                self.distance_types,
                [x.compute_distance(vals, ref_vals) for x in self.dist_classes],
            )
        )

        for idx in range(len(aggregate_ids)):
            idx_in_original_batch = select_idxs[idx]
            models = dict(
                zip(
                    self.model_names,
                    [
                        all_models[jdx][idx_in_original_batch]
                        for jdx in range(len(self.model_names))
                    ],
                )
            )
            features = dict(
                zip(
                    self.feature_names,
                    [
                        all_features[jdx][idx_in_original_batch]
                        for jdx in range(len(self.feature_names))
                    ],
                )
            )

            if len(self.log_writers) > 0:
                for k, distance_type in enumerate(self.distance_types):
                    self.log_writers[k].log(
                        {
                            "check": distances[distance_type][idx],
                            self.aggregate_measurable.feature_name: aggregate_ids[idx],  # type: ignore
                            self.count_measurable.feature_name: counts[idx],  # type: ignore
                            **models,
                            **features,
                        }
                    )
            else:
                for distance_type in self.distance_types:
                    plot_name = distance_type + "_" + str(self.reference)
                    self.log_handler.add_scalars(
                        self.dashboard_name + "_" + plot_name,
                        {"y_" + distance_type: distances[distance_type][idx]},
                        counts[idx],
                        self.dashboard_name,
                        features=features,
                        models=models,
                        file_name=str(aggregate_ids[idx]),
                    )

        # save the reference embeddings for use in the next batch
        self.cache.upsert_ids_n_col_values(
            np.asarray(list(ref_embs_prev.keys())),
            {"ref_embedding": np.asarray(list(ref_embs_prev.values()))},
        )



from abc import ABC
from typing import Any

import numpy as np


class AbstractDistance(ABC):
    def __init__(self) -> None:
        super().__init__()

    def compute_distance(self, base, reference) -> Any:
        raise Exception("Distance computation should be defined for each subclass")

    def check_compatibility(self, base, reference) -> None:
        if base.shape != reference.shape:
            raise Exception("Incompatible shapes for base and reference")



"""
Implement operators to evaluate question-response-context datapoints from a 
retrieval augmented pipeline. 
"""

from __future__ import annotations
import typing as t

from loguru import logger
import polars as pl

if t.TYPE_CHECKING:
    from uptrain.framework import Settings
from uptrain.operators.base import *


@register_op
class ResponseFactualScore(ColumnOp):
    """
    Grade how factual the generated response was.

     Attributes:
        col_question (str): Column name for the stored questions
        col_context: (str) Coloumn name for stored context
        col_response (str): Coloumn name for the stored responses

    Raises:
        Exception: Raises exception for any failed evaluation attempts



    """

    col_question: str = "question"
    col_context: str = "context"
    col_response: str = "response"
    col_out: str = "score_factual_accuracy"

    def setup(self, settings: t.Optional[Settings] = None):
        from uptrain.framework.remote import APIClient

        assert settings is not None
        self._api_client = APIClient(settings)
        return self

    def run(self, data: pl.DataFrame) -> TYPE_TABLE_OUTPUT:
        data_send = [
            {
                "question": row[self.col_question],
                "response": row[self.col_response],
                "context": row[self.col_context],
            }
            for row in data.to_dicts()
        ]
        try:
            results = self._api_client.evaluate("factual_accuracy", data_send)
        except Exception as e:
            logger.error(f"Failed to run evaluation for `ResponseFactualScore`: {e}")
            raise e
        
        assert results is not None
        return {"output": data.with_columns(pl.from_dicts(results).rename({"score_factual_accuracy": self.col_out}))}


@register_op
class ResponseCompleteness(ColumnOp):
    """
    Grade how complete the generated response was for the question specified.

    Attributes:
        col_question (str): Column name for the stored questions
        col_response (str): Coloumn name for the stored responses

    Raises:
        Exception: Raises exception for any failed evaluation attempts

    """

    col_question: str = "question"
    col_response: str = "response"
    col_out: str = "score_response_completeness"

    def setup(self, settings: t.Optional[Settings] = None):
        from uptrain.framework.remote import APIClient

        assert settings is not None
        self._api_client = APIClient(settings)
        return self

    def run(self, data: pl.DataFrame) -> TYPE_TABLE_OUTPUT:
        data_send = [
            {
                "question": row[self.col_question],
                "response": row[self.col_response],
            }
            for row in data.to_dicts()
        ]
        try:
            results = self._api_client.evaluate("response_completeness", data_send)
        except Exception as e:
            logger.error(f"Failed to run evaluation for `ResponseCompleteness`: {e}")
            raise e

        assert results is not None
        return {"output": data.with_columns(pl.from_dicts(results).rename({"score_response_completeness": self.col_out}))}


@register_op
class ResponseCompletenessWrtContext(ColumnOp):
    """
    Grade how complete the generated response was for the question specified given the information provided in the context.
    """

    col_question: str = "question"
    col_response: str = "response"
    col_context: str = "context"
    col_out: str = "score_response_completeness_wrt_context"

    def setup(self, settings: t.Optional[Settings] = None):
        from uptrain.framework.remote import APIClient

        assert settings is not None
        self._api_client = APIClient(settings)
        return self

    def run(self, data: pl.DataFrame) -> TYPE_TABLE_OUTPUT:
        data_send = [
            {
                "question": row[self.col_question],
                "response": row[self.col_response],
                "context": row[self.col_context],
            }
            for row in data.to_dicts()
        ]
        try:
            results = self._api_client.evaluate(
                "response_completeness_wrt_context", data_send
            )
        except Exception as e:
            logger.error(
                f"Failed to run evaluation for `ResponseCompletenessWrtContext`: {e}"
            )
            raise e

        assert results is not None
        return {"output": data.with_columns(pl.from_dicts(results).rename({"score_response_completeness_wrt_context": self.col_out}))}


@register_op
class ContextRelevance(ColumnOp):
    """
    Grade how relevant the context was to the question asked.

    Attributes:
        col_question: (str) Column Name for the stored questions
        col_context: (str) Coloumn name for stored context

    Raises:
        Exception: Raises exception for any failed evaluation attempts

    """

    col_question: str = "question"
    col_context: str = "context"
    col_out: str = "score_context_relevance"

    def setup(self, settings: t.Optional[Settings] = None):
        from uptrain.framework.remote import APIClient

        assert settings is not None
        self._api_client = APIClient(settings)
        return self

    def run(self, data: pl.DataFrame) -> TYPE_TABLE_OUTPUT:
        data_send = [
            {
                "question": row[self.col_question],
                "context": row[self.col_context],
            }
            for row in data.to_dicts()
        ]
        try:
            results = self._api_client.evaluate("context_relevance", data_send)
        except Exception as e:
            logger.error(f"Failed to run evaluation for `ContextRelevance`: {e}")
            raise e

        assert results is not None
        return {"output": data.with_columns(pl.from_dicts(results).rename({"score_context_relevance": self.col_out}))}


@register_op
class ResponseRelevance(ColumnOp):
    """
    Grades how relevant the generated response is or if it has any additional irrelevant information for the question asked.

    Attributes:
        col_question (str): Column name for the stored questions
        col_response (str): Coloumn name for the stored responses

    Raises:
        Exception: Raises exception for any failed evaluation attempts


    """

    col_question: str = "question"
    col_response: str = "response"
    col_out: str = "score_response_relevance"

    def setup(self, settings: t.Optional[Settings] = None):
        from uptrain.framework.remote import APIClient

        assert settings is not None
        self._api_client = APIClient(settings)
        return self

    def run(self, data: pl.DataFrame) -> TYPE_TABLE_OUTPUT:
        data_send = [
            {
                "question": row[self.col_question],
                "response": row[self.col_response],
            }
            for row in data.to_dicts()
        ]
        try:
            results = self._api_client.evaluate("response_relevance", data_send)
        except Exception as e:
            logger.error(f"Failed to run evaluation for `ResponseRelevance`: {e}")
            raise e

        assert results is not None
        return {"output": data.with_columns(pl.from_dicts(results).rename({"score_response_relevance": self.col_out}))}




@register_op
class ResponseConciseness(ColumnOp):
    """
    Grades how concise the generated response is or if it has any additional irrelevant information for the question asked.

    Attributes:
        col_question (str): Column name for the stored questions
        col_response (str): Coloumn name for the stored responses

    Raises:
        Exception: Raises exception for any failed evaluation attempts


    """

    col_question: str = "question"
    col_response: str = "response"
    col_out: str = "score_response_conciseness"

    def setup(self, settings: t.Optional[Settings] = None):
        from uptrain.framework.remote import APIClient

        assert settings is not None
        self._api_client = APIClient(settings)
        return self

    def run(self, data: pl.DataFrame) -> TYPE_TABLE_OUTPUT:
        data_send = [
            {
                "question": row[self.col_question],
                "response": row[self.col_response],
            }
            for row in data.to_dicts()
        ]
        try:
            results = self._api_client.evaluate("response_conciseness", data_send)
        except Exception as e:
            logger.error(f"Failed to run evaluation for `ResponseConciseness`: {e}")
            raise e

        assert results is not None
        return {"output": data.with_columns(pl.from_dicts(results).rename({"score_response_conciseness": self.col_out}))}




@register_op
class ResponseConsistency(ColumnOp):
    """
    Gives scores how relevant the generated response is for the question asked along with the context.

    Attributes:
        col_question (str): Column name for the stored questions
        col_response (str): Coloumn name for the stored responses
        col_context (str): Column name for the stored contexts

    Raises:
        Exception: Raises exception for any failed evaluation attempts


    """

    col_question: str = "question"
    col_response: str = "response"
    col_context: str = "context"
    col_out: str = "score_response_consistency"

    def setup(self, settings: t.Optional[Settings] = None):
        from uptrain.framework.remote import APIClient

        assert settings is not None
        self._api_client = APIClient(settings)
        return self

    def run(self, data: pl.DataFrame) -> TYPE_TABLE_OUTPUT:
        data_send = [
            {
                "question": row[self.col_question],
                "response": row[self.col_response],
                "context": row[self.col_context]
            }
            for row in data.to_dicts()
        ]
        try:
            results = self._api_client.evaluate("response_consistency", data_send)
        except Exception as e:
            logger.error(f"Failed to run evaluation for `ResponseConsistency`: {e}")
            raise e

        assert results is not None
        return {"output": data.with_columns(pl.from_dicts(results).rename({"score_response_consistency": self.col_out}))}



"""
Implement checks to test if a piece of text has been taken from a source.

This module provides the `METEORScore` class, which allows comparing a generated text with a source text to check if it has been taken from the source. 
It uses the METEOR score metric, specifically the METEOR score, to measure the similarity between the two texts.
Metric for Evaluation of Translation with Explicit Ordering (METEOR) score is a metric that measures the quality 
of generated text based on the alignment between the generated text and the reference text. 

"""

from __future__ import annotations
import typing as t

from loguru import logger
import polars as pl
from uptrain.framework import Settings

if t.TYPE_CHECKING:
    from uptrain.framework import Settings
from uptrain.operators.base import *
from uptrain.utilities import lazy_load_dep

nltk = lazy_load_dep("nltk", "nltk")


@register_op
class METEORScore(ColumnOp):
    """
    Operator to compare a generated text with a source text using the METEOR score metric.
    More about METEOR - https://en.wikipedia.org/wiki/METEOR

    Attributes:
        col_in_generated (str): The name of the input column containing the generated text.
        col_in_source (str): The name of the input column containing the source text.
        col_out (str): The name of the output column containing the METEOR scores.

    Returns:
        dict: A dictionary containing the METEOR scores for each pair of generated and source text.

    Example:
        ```
        import polars as pl
        from uptrain.operators import METEORScore

        # Create a DataFrame
        df = pl.DataFrame({
            "text_generated": ["The cat was sat on the mat"],
            "text_source": ["The cat sat on the mat"]
        })

        # Create an instance of the METEORScore class
        METEOR_op = METEORScore()

        # Calculate the METEOR scores
        scores = METEOR_op.run(df)["output"]

        # Print the METEOR scores
        print(scores["METEOR_score"])
        ```

    Output:
        ```
        shape: (1,)
        Series: 'METEOR_score' [i64]
        [
            96
        ]
        ```

    """

    col_in_generated: str = "text_generated"
    col_in_source: str = "text_source"
    col_out: str = "METEOR_score"

    def setup(self, settings: Settings):
        return self

    def run(self, data: pl.DataFrame) -> TYPE_TABLE_OUTPUT:
        text_generated = data.get_column(self.col_in_generated)  # candidate/preds
        text_source = data.get_column(self.col_in_source)  # reference/target

        results = []
        scores = []
        for i in range(len(text_generated)):
            scorer = nltk.translate.meteor
            if text_source[i] is None or text_generated[i] is None:
                scores.append({"METEOR": 0})
            else:
                preds = nltk.word_tokenize(text_generated[i])
                target = [nltk.word_tokenize(text_source[i])]
                scores.append({"METEOR": scorer(references=target, hypothesis=preds)})

        results = pl.Series([int(x["METEOR"] * 100) for x in scores])
        return {"output": data.with_columns([results.alias(self.col_out)])}



import numpy as np

from typing import List, Union

from uptrain.v0.core.classes.distances import AbstractDistance


class CosineDistance(AbstractDistance):
    """Class that computes cosine distance between base and reference vectors."""

    def calculate_norm(self, vector) -> float:
        if len(vector.shape) > 1:
            return np.linalg.norm(vector, axis=1)
        return np.abs(vector)

    def compute_distance(
        self, base: Union[List, np.ndarray], reference: Union[List, np.ndarray]
    ) -> np.ndarray:
        """Compute the cosine distance between base and reference.

        The cosine distance between vectors x and y is defined as:
            1 - cosine_similarity(x, y)
        The cosine similarity can be computed as:
            dot_product(x, y) / (normalized(x) * normalized(y))

        For further reading, check [this](https://en.wikipedia.org/wiki/Cosine_similarity) out.

        Parameters
        ----------
        base
            x in the above equations.
            It is the base vector for cosine distance computation.
        reference
            y in the above equations.
            It is the reference vector for cosine distance computation.

        Returns
        -------
        np.ndarray
            The cosine distance between base and reference.
        """
        base = np.array(base)
        reference = np.array(reference)

        self.check_compatibility(base, reference)

        base_norm = self.calculate_norm(base)
        ref_norm = self.calculate_norm(reference)

        return np.array(
            [
                1 - np.dot(base[i], reference[i]) / (base_norm[i] * ref_norm[i])
                for i in range(base.shape[0])
            ]
        )



import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import uptrain.v0 as v0


def test_concept_drift_adwin():
    """Test concept drift detection using ADWIN algorithm.

    We generate 2 distributions:
        - The first distribution contains 1000 values with mean 0.2 and standard deviation 0.03
        - The second distribution contains 1000 values with mean 0.5 and standard deviation 0.08

    Let's assume the following as well:
        For the first distribution:
            - ground truth is 0.2
        For the second distribution
            - ground truth is 0.5 for the first half of values
            - ground truth is 0.6 for the first half of values

    To maintain the simplicity of this test, we will also assume that we have a hypothetical
    trained model that has learned to predict the output as the same as input. However, as can
    be seen from the ground truth values, this is not the case. The model will correctly predict
    the first distribution as well as the first half of the second distribution. However, it
    will make incorrect predictions on the second half of the second distribution.

    This means that between the model predictions and ground truths, a concept drift will arise.
    The drift will arise somewhere around the 1500th value.

    Why?
      - The model will predict values approximately equal to 0.2 for the first distribution.
        This is correctly predicted.
      - The model will predict values approximately equal to 0.5 for the first half of the
        second distribution. This is correctly predicted as well.
      - The model will predict values approximately equal to 0.5 for the second half of the
        second distribution. This is incorrectly predicted, which causes the error rate to
        grow higher. After it crosses a set threshold, a concept drift alert is raised.

    In this test, we use the UpTrain framework to log the inputs, outputs, and ground truths.
    We set up the framework to use the ADWIN algorithm to monitor the error rate of the model.
    We use two error rates for detection of concept drift:
        - The first error rate is calculated as the mean absolute difference between the
          predictions and the ground truths.
        - The second error rate is calculated as the mean absolute percentage difference
          between the predictions and the ground truths.

    The test passes if the ADWIN algorithm successfully detects the concept drift and raises
    an alert around the point where the model's predictions start becoming inaccurate.
    """

    random_state = np.random.RandomState(seed=1337)
    n = 1000
    params = [(0.2, 0.03, n), (0.5, 0.08, n)]
    distributions = np.array([random_state.normal(*param) for param in params])

    ground_truths = np.array([0.2] * n + [0.5] * (n // 2) + [0.6] * (n // 2))
    stream = distributions.flatten()

    def plot_data(drifts=None):
        """Helper function to plot data."""

        _ = plt.figure(figsize=(7, 3), tight_layout=True)
        gridspecs = gridspec.GridSpec(1, 2, width_ratios=[3, 1])

        ax1, ax2 = plt.subplot(gridspecs[0]), plt.subplot(gridspecs[1])
        ax1.grid()
        ax1.plot(stream, label="stream")
        ax2.grid(axis="y")

        for index, dist in enumerate(distributions):
            ax2.hist(dist, label=f"dist_{index}")

        drift_color = "red"
        if drifts is not None:
            for drift_detected in drifts:
                ax1.axvline(drift_detected, color=drift_color)
        plt.tight_layout()
        plt.show()

    # Plot data for debugging
    # plot_data()

    # Create a configuration for the framework
    cfg = {
        # Define checks to be performed
        "checks": [
            # First check: monitor concept drift using ADWIN algorithm with
            # mean absolute error as the measurable for the feature "data".
            {
                "type": v0.Monitor.CONCEPT_DRIFT,
                "algorithm": v0.DataDriftAlgo.ADWIN,
                "measurable_args": {"type": v0.MeasurableType.MAE},
                # Configurable parameters that the ADWIN algorithm supports
                "delta": 0.002,
                "clock": 32,
                "max_buckets": 5,
                "min_window_length": 5,
                "grace_period": 5,
            },
            # Second check: monitor concept drift using ADWIN algorithm with
            # mean absolute percentage error as the measurable for the feature "data".
            # Notice that we need a different delta value (significance value)
            # for using MAPE as the measurable instead of MAE.
            {
                "type": v0.Monitor.CONCEPT_DRIFT,
                "algorithm": v0.DataDriftAlgo.ADWIN,
                "measurable_args": {
                    "type": v0.MeasurableType.MAPE,
                    "feature_name": "data",
                },
                "delta": 0.1,
            },
        ],
        # Specify where the logging data should be stored
        "retraining_folder": "uptraining_smart_data_concept_drift_adwin",
        # Specify logging arguments
        # "st_logging" should be True if we want streamlit logging, False otherwise
        "logging_args": {"st_logging": True},
    }

    framework = v0.Framework(cfg)
    batch_size = 64
    size = len(stream)

    # Feed the data to the framework
    for i in range(size // batch_size):
        inputs = {"data": stream[i * batch_size : (i + 1) * batch_size]}
        outputs = stream[i * batch_size : (i + 1) * batch_size]
        gts = ground_truths[i * batch_size : (i + 1) * batch_size]
        ids = framework.log(inputs=inputs, outputs=outputs)
        framework.log(identifiers=ids, gts=gts)



import numpy as np

from typing import Any

from uptrain.v0.core.lib.helper_funcs import (
    extract_data_points_from_batch,
    combine_data_points_for_batch,
)
from uptrain.v0.core.classes.measurables import AbstractMeasurable


class Measurable(AbstractMeasurable):
    """Class that contains the core logic for computation/logging of a measurable.

    This class is a subclass of AbstractMeasurable and implements the computation
    and logging of a measurable or evaluatable object. It can optionally use a
    cache to avoid recomputing values that have already been computed.
    """

    def __init__(self, framework) -> None:
        super().__init__()
        self.framework = framework
        self.cache_present_or_not = None

    def compute_and_log(self, inputs=None, outputs=None, gts=None, extra=None) -> Any:
        """Computes the measurement and logs the result.

        This method computes the measurement of this object using the overriden
        _compute method, and logs the result. If caching is enabled in the
        UpTrain framework object, this method may use the cache to avoid recomputing
        values that have already been computed.

        Parameters
        ----------
        inputs
            Inputs values to use in the computation
        outputs
            Outputs values to use in the computation
        gts
            Ground truth values to use in the computation
        extra
            Additional information to use in the computation

        Returns
        -------
        Any
            Result of the computation
        """

        if self.framework.use_cache:
            if self.cache_present_or_not is None:
                self.cache_present_or_not = 1
                self.framework.cache.update({self.col_name(): {}})

            compute_masks = []
            compute_idxs = []
            non_compute_vals = []
            for idx in range(len(extra["id"])):
                id = extra["id"][idx]
                if id in self.framework.cache[self.col_name()]:
                    compute_masks.append(0)
                    non_compute_vals.append(self.framework.cache[self.col_name()][id])
                else:
                    compute_masks.append(1)
                    compute_idxs.append(idx)

            compute_masks = np.array(compute_masks)
            compute_idxs = np.array(compute_idxs)

            compute_vals = self._compute(
                inputs=extract_data_points_from_batch(inputs, compute_idxs),
                outputs=extract_data_points_from_batch(outputs, compute_idxs),
                gts=extract_data_points_from_batch(gts, compute_idxs),
                extra=extract_data_points_from_batch(extra, compute_idxs),
            )

            # TODO: Replace below for loop by numpy operations
            vals = []
            compute_offset = 0
            non_compute_offset = 0
            for idx in range(len(extra["id"])):
                if compute_masks[idx] == 1:
                    vals.append(
                        extract_data_points_from_batch(compute_vals, [compute_offset])
                    )
                    compute_offset += 1
                else:
                    vals.append(
                        extract_data_points_from_batch(
                            non_compute_vals, [non_compute_offset]
                        )
                    )
                    non_compute_offset += 1
            vals = combine_data_points_for_batch(vals)
            try:
                self._log(extra["id"], vals)
            except:
                raise("Error in logging")
        else:
            vals = self._compute(inputs=inputs, outputs=outputs, gts=gts, extra=extra)
        return vals

    def _log(self, ids: np.ndarray, vals: np.ndarray) -> None:
        """Logs the result of the computation.

        This method logs the result of the computation for the given ids and vals
        using the log_measurable method of UpTrain framework object.

        Parameters
        ----------
        ids
            List of unique identifiers for inputs, which is used to match them with
            their computed values in the framework's logging system
        vals
            List of computed measurable values for inputs
        """
        self.framework.log_measurable(ids, vals, self.col_name())



import importlib.util
import os
import threading
import typing as t

from loguru import logger


class StreamlitRunner:
    log_folder: str
    launch_cmd: str

    def __init__(self, log_folder: str, port: t.Optional[int] = None):
        self.log_folder = log_folder
        if "config.json" not in os.listdir(self.log_folder):
            logger.warning("No Uptrain config found in the specified log folder.")

        path_uptrain_init: str = importlib.util.find_spec("uptrain").origin  # type: ignore
        path_st_run = os.path.join(
            os.path.dirname(path_uptrain_init), "dashboard", "st_run.py"
        )
        port_arg = "" if port is None else f"--server.port {port}"
        nowatch_arg = "--server.fileWatcherType none"
        self.launch_cmd = (
            f"streamlit run {path_st_run} {port_arg} {nowatch_arg} -- {self.log_folder}"
        )
        logger.info(f"Streamlit launch command: {self.launch_cmd}")

    def start(self):
        t = threading.Thread(target=lambda: os.system(self.launch_cmd), args=([]))
        t.start()
